import asyncio
import base64
import hmac
import math
import secrets
import sqlite3
import time
from enum import Enum, IntEnum, auto
from typing import TYPE_CHECKING, Literal, Optional, Protocol, Type, Union, cast

from lonelypsp.auth.config import (
    AuthResult,
    ToBroadcasterAuthConfig,
    ToSubscriberAuthConfig,
)
from lonelypsp.auth.set_subscriptions_info import SetSubscriptionsInfo
from lonelypsp.compat import fast_dataclass
from lonelypsp.stateful.messages.configure import S2B_Configure
from lonelypsp.stateful.messages.confirm_configure import B2S_ConfirmConfigure
from lonelypsp.stateful.messages.continue_receive import S2B_ContinueReceive
from lonelypsp.stateful.messages.disable_zstd_custom import B2S_DisableZstdCustom
from lonelypsp.stateful.messages.enable_zstd_custom import B2S_EnableZstdCustom
from lonelypsp.stateful.messages.enable_zstd_preset import B2S_EnableZstdPreset
from lonelypsp.stateless.make_strong_etag import StrongEtag
from lonelypsp.util.cancel_and_check import cancel_and_check

if TYPE_CHECKING:
    from lonelypsp.auth.config import (
        ToBroadcasterAuthConfig,
        ToSubscriberAuthConfig,
    )


class IncomingHmacAuthDBConfig(Protocol):
    """Describes how to verify that incoming hash-based message authentication
    codes are not being reused.
    """

    async def setup_hmac_auth_db(self) -> None: ...
    async def teardown_hmac_auth_db(self) -> None: ...

    async def mark_code_used(self, /, *, code: bytes) -> Literal["conflict", "ok"]: ...


class IncomingHmacAuthDBReentrantConfig:
    """Delegates to another implementation without forwarding all but the
    outermost setup/teardown, making it reentrant
    """

    def __init__(self, delegate: IncomingHmacAuthDBConfig) -> None:
        self.delegate = delegate
        self.depth = 0
        self.lock = asyncio.Lock()

    async def setup_hmac_auth_db(self) -> None:
        async with self.lock:
            if self.depth == 0:
                await self.delegate.setup_hmac_auth_db()
            self.depth += 1

    async def teardown_hmac_auth_db(self) -> None:
        async with self.lock:
            if self.depth <= 0:
                return

            self.depth -= 1
            if self.depth == 0:
                await self.delegate.teardown_hmac_auth_db()

    async def mark_code_used(self, /, *, code: bytes) -> Literal["conflict", "ok"]:
        return await self.delegate.mark_code_used(code=code)


class IncomingHmacAuthNoneDBConfig:
    """Does not store recent tokens and thus cannot check if they've been recently
    used. Technically, this is vulnerable to replay attacks, though the scope is
    rather limited.
    """

    async def setup_hmac_auth_db(self) -> None: ...
    async def teardown_hmac_auth_db(self) -> None: ...

    async def mark_code_used(self, /, *, code: bytes) -> Literal["conflict", "ok"]:
        return "ok"


if TYPE_CHECKING:
    _: Type[IncomingHmacAuthDBConfig] = IncomingHmacAuthNoneDBConfig


class IncomingHmacAuthSqliteDBConfig:
    """Stores recent tokens in a sqlite database, cleaning them in the background
    occassionally. This is only effective if there is only one broadcaster, though
    it will help detect replay attacks even if there are multiple broadcasters.
    """

    def __init__(
        self,
        database: str,
        *,
        token_lifetime: int = 180,
        cleanup_batch_delay: float = 10.0,
    ) -> None:
        self.database = database
        """The database url. You can pass `:memory:` to create a SQLite database that
        exists only in memory, otherwise, this is typically the path to a sqlite file
        (usually has the `db` extension).
        """

        self.token_lifetime: int = token_lifetime
        """The minimum time in seconds before we forget about a token, must be at
        least as long as the token is accepted for this to be effective at preventing
        replay attacks.
        """

        self.cleanup_batch_delay: float = cleanup_batch_delay
        """The minimum time in seconds between cleaning up tokens that have expired"""

        self.conn: Optional[sqlite3.Connection] = None
        """The connection to the database"""

        self.cursor: Optional[sqlite3.Cursor] = None
        """The cursor that can be used iff you will be done using it before yielding
        to the event loop
        """

        self.background_task: Optional[asyncio.Task[None]] = None
        """The cleanup task that runs in the background"""

        self.cleanup_wakeup: asyncio.Event = asyncio.Event()
        """An event that is set whenever a token is created and will be waited on
        when there are no tokens in the store
        """

    async def setup_hmac_auth_db(self) -> None:
        assert self.background_task is None, "already entered, not re-entrant"
        conn = sqlite3.connect(self.database, isolation_level=None)
        try:
            cursor = conn.cursor()
            try:
                cursor.execute("BEGIN IMMEDIATE TRANSACTION")
                cursor.execute(
                    "CREATE TABLE IF NOT EXISTS httppubsub_hmacs (code BLOB PRIMARY KEY, expires_at INTEGER NOT NULL) WITHOUT ROWID"
                )
                cursor.execute(
                    "CREATE INDEX IF NOT EXISTS idx_httppubsub_hmacs_expires_at ON httppubsub_hmacs (expires_at)"
                )
                cursor.execute("COMMIT")
                self.conn = conn
                self.cursor = cursor
                self.background_task = asyncio.create_task(self._cleanup_codes())
            except BaseException:
                cursor.close()
                raise
        except BaseException:
            conn.close()
            raise

    async def teardown_hmac_auth_db(self) -> None:
        task = self.background_task
        conn = self.conn
        cursor = self.cursor
        self.background_task = None
        self.conn = None
        self.cursor = None
        try:
            if task is not None:
                await cancel_and_check(task)
        finally:
            try:
                if cursor is not None:
                    cursor.close()
            finally:
                if conn is not None:
                    conn.close()

    async def mark_code_used(self, /, *, code: bytes) -> Literal["conflict", "ok"]:
        assert self.conn is not None and self.cursor is not None, "not entered"
        self.cursor.execute("BEGIN IMMEDIATE TRANSACTION")
        try:
            self.cursor.execute(
                "INSERT INTO httppubsub_hmacs (code, expires_at) "
                "SELECT ?, ? "
                "WHERE"
                " NOT EXISTS ("
                "SELECT 1 FROM httppubsub_hmacs WHERE code = ?"
                ")",
                (code, math.ceil(time.time() + self.token_lifetime), code),
            )
            is_conflict = self.cursor.rowcount == 0
            self.cursor.execute("COMMIT")
        except BaseException:
            self.cursor.execute("ROLLBACK")
            raise
        if not is_conflict:
            self.cleanup_wakeup.set()
        return "conflict" if is_conflict else "ok"

    async def _cleanup_codes(self) -> None:
        assert self.conn is not None and self.cursor is not None, "not entered"
        cursor = self.cursor

        while True:
            now = time.time()
            cursor.execute("BEGIN IMMEDIATE TRANSACTION")
            try:
                cursor.execute(
                    "DELETE FROM httppubsub_hmacs WHERE expires_at < ?",
                    (math.floor(now),),
                )
                cursor.execute(
                    "SELECT expires_at FROM httppubsub_hmacs ORDER BY expires_at ASC LIMIT 1"
                )
                next_expires_at = cast(Optional[int], cursor.fetchone())
                cursor.execute("COMMIT")
            except BaseException:
                cursor.execute("ROLLBACK")
                raise

            if next_expires_at is None:
                self.cleanup_wakeup.clear()
                await self.cleanup_wakeup.wait()
                await asyncio.sleep(self.token_lifetime + self.cleanup_batch_delay)
                continue

            await asyncio.sleep(max(next_expires_at - now, self.cleanup_batch_delay))


class TokenInfoType(Enum):
    """What we can discover when extracting token information"""

    UNAUTHORIZED = auto()
    """the information was missing"""
    FORBIDDEN = auto()
    """the information was malformed or definitely not acceptable"""
    FOUND = auto()
    """the information was found but not checked"""


@fast_dataclass
class TokenInfoUnauthorized:
    type: Literal[TokenInfoType.UNAUTHORIZED]


@fast_dataclass
class TokenInfoForbidden:
    type: Literal[TokenInfoType.FORBIDDEN]


@fast_dataclass
class TokenInfoFound:
    type: Literal[TokenInfoType.FOUND]
    timestamp: int
    """approximately when the token was created"""
    nonce: str
    """a unique value to avoid ever resigning the exact same data"""
    hmac: bytes
    """the provided hash-based message authentication code"""


def get_token(
    authorization: Optional[str], /, *, now: float, token_lifetime: float
) -> Union[TokenInfoUnauthorized, TokenInfoForbidden, TokenInfoFound]:
    """Extracts the hash-based message authentication code from the authorization
    header and performs sanity checks that don't depend on the data being
    signed or the database
    """
    if authorization is None:
        return TokenInfoUnauthorized(type=TokenInfoType.UNAUTHORIZED)

    if not authorization.startswith("X-HMAC "):
        return TokenInfoForbidden(type=TokenInfoType.FORBIDDEN)

    timestamp_nonce_and_token = authorization[len("X-HMAC ") :]
    sep_index = timestamp_nonce_and_token.find(":")
    if sep_index == -1:
        return TokenInfoForbidden(type=TokenInfoType.FORBIDDEN)

    timestamp_str = timestamp_nonce_and_token[:sep_index]
    try:
        timestamp = int(timestamp_str)
    except ValueError:
        return TokenInfoForbidden(type=TokenInfoType.FORBIDDEN)

    # clock drift means the time could be in the future
    if abs(now - timestamp) > token_lifetime:
        return TokenInfoForbidden(type=TokenInfoType.FORBIDDEN)

    nonce_and_token = timestamp_nonce_and_token[sep_index + 1 :]
    sep_index = nonce_and_token.find(":")
    if sep_index == -1:
        return TokenInfoForbidden(type=TokenInfoType.FORBIDDEN)

    nonce = nonce_and_token[:sep_index]

    code_str = nonce_and_token[sep_index + 1 :]
    try:
        code = base64.b64decode(code_str + "==")
    except ValueError:
        return TokenInfoForbidden(type=TokenInfoType.FORBIDDEN)

    if len(code) != 64:
        return TokenInfoForbidden(type=TokenInfoType.FORBIDDEN)

    return TokenInfoFound(
        type=TokenInfoType.FOUND, timestamp=timestamp, nonce=nonce, hmac=code
    )


def make_nonce() -> str:
    """standard way to generate a nonce for HMAC tokens"""
    return secrets.token_urlsafe(4)


async def check_code(
    *, secret: bytes, to_sign: bytes, code: bytes, db: IncomingHmacAuthDBConfig
) -> AuthResult:
    """Checks if the provided hash-based message authentication code matches
    whats expected given it should have signed `to_sign` with the given
    secret and should not be a replay of a recent token (via db)
    """
    expected_hmac = hmac.new(secret, to_sign, "sha512").digest()
    if not hmac.compare_digest(code, expected_hmac):
        return AuthResult.FORBIDDEN

    if await db.mark_code_used(code=code) == "conflict":
        return AuthResult.FORBIDDEN

    return AuthResult.OK


def sign(*, secret: bytes, to_sign: bytes, nonce: str, now: float) -> str:
    """Generates the hash-based message authentication code for the given data then
    formats it in the standard way for transfer
    """
    hmac_token = hmac.new(secret, to_sign, "sha512").digest()
    return f"X-HMAC {int(now)}:{nonce}:{base64.b64encode(hmac_token).decode('ascii')}"


class AuthMessageType(IntEnum):
    """In order to protect against a man in the middle which convinces
    a trusted party to sign a token with arbitrary data that makes it look
    like a different type of message, all messages to sign are prefixed with
    an int specifying how they are intending to be parsed so that confusion
    will always result in a wrong digest

    Example of this attack:

    suppose there are two messages, one which is:
    - 2 bytes: length of url (N)
    - N bytes: url
    - 2 bytes: length of topic (M)
    - M bytes: topic

    and another message which is just
    - 2 bytes: length of url (N)
    - N bytes: url

    if the attacker can convince the subscriber to sign the second message
    where the url is specifically crafted to look like a url prefix followed
    by 2 utf-8 bytes that look like the length of the remaining topic, then
    a valid topic, then the attacker can use that signature and replace the body
    to get a different result
    """

    SUBSCRIBE_EXACT = auto()
    SUBSCRIBE_GLOB = auto()
    NOTIFY = auto()
    STATEFUL_CONFIGURE = auto()
    CHECK_SUBSCRIPTIONS = auto()
    SET_SUBSCRIPTIONS = auto()
    STATEFUL_CONTINUE_RECEIVE = auto()
    CONFIRM_RECEIVE = auto()
    CONFIRM_MISSED = auto()
    RECEIVE = auto()
    MISSED = auto()
    CONFIRM_SUBSCRIBE_EXACT = auto()
    CONFIRM_SUBSCRIBE_GLOB = auto()
    CONFIRM_NOTIFY = auto()
    STATEFUL_CONFIRM_CONFIGURE = auto()
    STATEFUL_ENABLE_ZSTD_PRESET = auto()
    STATEFUL_ENABLE_ZSTD_CUSTOM = auto()
    STATEFUL_DISABLE_ZSTD_CUSTOM = auto()


class ToBroadcasterHmacAuth:
    """Sets up and allows requests to the broadcaster which include hash-based
    message authentication codes. These tokens can only be generated if the
    sender knows the shared secret.

    The authorization header is formatted as follows:
    `X-HMAC <timestamp>:<nonce>:<token>`, where timestamp is integer seconds
    from the epoch, the nonce is to ensure uniqueness, and the token is the HMAC
    of the relevant information signed with the shared secret.

    This is a secure way to verify requests even if the underlying message
    and headers are not encrypted, as the shared secret cannot be discovered
    by an attacker who can only see the messages. However, it will not do
    anything to prevent the contents of the messages from being read

    Generally, this is an appropriate defense-in-depth measure to use in
    internal networks where you cannot setup TLS. It is also effective when TLS
    is available, though token authorization will require less CPU time for both
    broadcasters and subscribers and is secure if headers are encrypted.
    """

    def __init__(
        self,
        *,
        secret: str,
        token_lifetime: float = 120,
        db_config: IncomingHmacAuthDBConfig,
    ) -> None:
        self.secret = base64.urlsafe_b64decode(secret + "==")
        """The shared secret used to generate the HMAC tokens"""
        assert len(self.secret) == 64, "subscriber_secret must be 64 bytes long"
        self.token_lifetime = token_lifetime
        """How long after a token is created that we still accept it"""
        self.db_config = db_config
        """The configuration for the recent codes database"""

    async def setup_to_broadcaster_auth(self) -> None:
        await self.db_config.setup_hmac_auth_db()

    async def teardown_to_broadcaster_auth(self) -> None:
        await self.db_config.teardown_hmac_auth_db()

    def _prepare_subscribe_exact(
        self,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        exact: bytes,
        timestamp: int,
        nonce: str,
    ) -> bytes:
        encoded_url = url.encode("utf-8")
        encoded_recovery = b"" if recovery is None else recovery.encode("utf-8")
        encoded_timestamp = timestamp.to_bytes(8, "big")
        encoded_nonce = nonce.encode("utf-8")

        return b"".join(
            [
                int(AuthMessageType.SUBSCRIBE_EXACT).to_bytes(1, "big"),
                encoded_timestamp,
                len(encoded_nonce).to_bytes(1, "big"),
                encoded_nonce,
                len(tracing).to_bytes(2, "big"),
                tracing,
                len(encoded_url).to_bytes(2, "big"),
                encoded_url,
                len(encoded_recovery).to_bytes(2, "big"),
                encoded_recovery,
                len(exact).to_bytes(2, "big"),
                exact,
            ]
        )

    async def authorize_subscribe_exact(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        exact: bytes,
        now: float,
    ) -> Optional[str]:
        nonce = make_nonce()

        to_sign = self._prepare_subscribe_exact(
            tracing=tracing,
            url=url,
            recovery=recovery,
            exact=exact,
            timestamp=int(now),
            nonce=nonce,
        )
        return sign(secret=self.secret, to_sign=to_sign, nonce=nonce, now=now)

    async def is_subscribe_exact_allowed(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        exact: bytes,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        token = get_token(authorization, now=now, token_lifetime=self.token_lifetime)
        if token.type == TokenInfoType.UNAUTHORIZED:
            return AuthResult.UNAUTHORIZED
        if token.type == TokenInfoType.FORBIDDEN:
            return AuthResult.FORBIDDEN

        to_sign = self._prepare_subscribe_exact(
            tracing=tracing,
            url=url,
            exact=exact,
            recovery=recovery,
            timestamp=token.timestamp,
            nonce=token.nonce,
        )
        return await check_code(
            secret=self.secret, to_sign=to_sign, code=token.hmac, db=self.db_config
        )

    def _prepare_subscribe_glob(
        self,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        glob: str,
        timestamp: int,
        nonce: str,
    ) -> bytes:
        encoded_url = url.encode("utf-8")
        encoded_recovery = b"" if recovery is None else recovery.encode("utf-8")
        encoded_timestamp = timestamp.to_bytes(8, "big")
        encoded_glob = glob.encode("utf-8")
        encoded_nonce = nonce.encode("utf-8")

        return b"".join(
            [
                int(AuthMessageType.SUBSCRIBE_GLOB).to_bytes(1, "big"),
                encoded_timestamp,
                len(encoded_nonce).to_bytes(1, "big"),
                encoded_nonce,
                len(tracing).to_bytes(2, "big"),
                tracing,
                len(encoded_url).to_bytes(2, "big"),
                encoded_url,
                len(encoded_recovery).to_bytes(2, "big"),
                encoded_recovery,
                len(encoded_glob).to_bytes(2, "big"),
                encoded_glob,
            ]
        )

    async def authorize_subscribe_glob(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        glob: str,
        now: float,
    ) -> Optional[str]:
        nonce = make_nonce()
        to_sign = self._prepare_subscribe_glob(
            tracing=tracing,
            url=url,
            recovery=recovery,
            glob=glob,
            timestamp=int(now),
            nonce=nonce,
        )
        return sign(secret=self.secret, to_sign=to_sign, nonce=nonce, now=now)

    async def is_subscribe_glob_allowed(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        glob: str,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        token = get_token(authorization, now=now, token_lifetime=self.token_lifetime)
        if token.type == TokenInfoType.UNAUTHORIZED:
            return AuthResult.UNAUTHORIZED
        if token.type == TokenInfoType.FORBIDDEN:
            return AuthResult.FORBIDDEN

        to_sign = self._prepare_subscribe_glob(
            tracing=tracing,
            url=url,
            glob=glob,
            recovery=recovery,
            timestamp=token.timestamp,
            nonce=token.nonce,
        )
        return await check_code(
            secret=self.secret, to_sign=to_sign, code=token.hmac, db=self.db_config
        )

    def _prepare_notify(
        self,
        tracing: bytes,
        topic: bytes,
        identifier: bytes,
        message_sha512: bytes,
        timestamp: int,
        nonce: str,
    ) -> bytes:
        encoded_timestamp = timestamp.to_bytes(8, "big")
        encoded_nonce = nonce.encode("utf-8")

        return b"".join(
            [
                int(AuthMessageType.NOTIFY).to_bytes(1, "big"),
                encoded_timestamp,
                len(encoded_nonce).to_bytes(1, "big"),
                encoded_nonce,
                len(tracing).to_bytes(2, "big"),
                tracing,
                len(topic).to_bytes(2, "big"),
                topic,
                len(identifier).to_bytes(1, "big"),
                identifier,
                message_sha512,
            ]
        )

    async def authorize_notify(
        self,
        /,
        *,
        tracing: bytes,
        topic: bytes,
        identifier: bytes,
        message_sha512: bytes,
        now: float,
    ) -> Optional[str]:
        nonce = make_nonce()
        to_sign = self._prepare_notify(
            tracing=tracing,
            topic=topic,
            identifier=identifier,
            message_sha512=message_sha512,
            timestamp=int(now),
            nonce=nonce,
        )
        return sign(secret=self.secret, to_sign=to_sign, nonce=nonce, now=now)

    async def is_notify_allowed(
        self,
        /,
        *,
        tracing: bytes,
        topic: bytes,
        identifier: bytes,
        message_sha512: bytes,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        assert len(message_sha512) == 64, "message_sha512 must be 64 bytes long"
        token = get_token(authorization, now=now, token_lifetime=self.token_lifetime)
        if token.type == TokenInfoType.UNAUTHORIZED:
            return AuthResult.UNAUTHORIZED
        if token.type == TokenInfoType.FORBIDDEN:
            return AuthResult.FORBIDDEN

        to_sign = self._prepare_notify(
            tracing=tracing,
            topic=topic,
            identifier=identifier,
            message_sha512=message_sha512,
            timestamp=token.timestamp,
            nonce=token.nonce,
        )
        return await check_code(
            secret=self.secret, to_sign=to_sign, code=token.hmac, db=self.db_config
        )

    def _prepare_stateful_configure(
        self,
        /,
        *,
        tracing: bytes,
        subscriber_nonce: bytes,
        enable_zstd: bool,
        enable_training: bool,
        initial_dict: int,
        timestamp: int,
        nonce: str,
    ) -> bytes:
        encoded_timestamp = timestamp.to_bytes(8, "big")
        encoded_nonce = nonce.encode("utf-8")

        return b"".join(
            [
                int(AuthMessageType.STATEFUL_CONFIGURE).to_bytes(1, "big"),
                encoded_timestamp,
                len(encoded_nonce).to_bytes(1, "big"),
                encoded_nonce,
                len(tracing).to_bytes(2, "big"),
                tracing,
                len(subscriber_nonce).to_bytes(1, "big"),
                subscriber_nonce,
                b"\1" if enable_zstd else b"\0",
                b"\1" if enable_training else b"\0",
                initial_dict.to_bytes(2, "big"),
            ]
        )

    async def authorize_stateful_configure(
        self,
        /,
        *,
        tracing: bytes,
        subscriber_nonce: bytes,
        enable_zstd: bool,
        enable_training: bool,
        initial_dict: int,
    ) -> Optional[str]:
        nonce = make_nonce()
        to_sign = self._prepare_stateful_configure(
            tracing=tracing,
            subscriber_nonce=subscriber_nonce,
            enable_zstd=enable_zstd,
            enable_training=enable_training,
            initial_dict=initial_dict,
            timestamp=int(time.time()),
            nonce=nonce,
        )
        return sign(secret=self.secret, to_sign=to_sign, nonce=nonce, now=time.time())

    async def is_stateful_configure_allowed(
        self, /, *, message: S2B_Configure, now: float
    ) -> AuthResult:
        token = get_token(
            message.authorization, now=now, token_lifetime=self.token_lifetime
        )
        if token.type == TokenInfoType.UNAUTHORIZED:
            return AuthResult.UNAUTHORIZED
        if token.type == TokenInfoType.FORBIDDEN:
            return AuthResult.FORBIDDEN

        to_sign = self._prepare_stateful_configure(
            tracing=message.tracing,
            subscriber_nonce=message.subscriber_nonce,
            enable_zstd=message.enable_zstd,
            enable_training=message.enable_training,
            initial_dict=message.initial_dict,
            timestamp=token.timestamp,
            nonce=token.nonce,
        )
        return await check_code(
            secret=self.secret, to_sign=to_sign, code=token.hmac, db=self.db_config
        )

    def _prepare_check_subscriptions(
        self, tracing: bytes, url: str, timestamp: int, nonce: str
    ) -> bytes:
        encoded_url = url.encode("utf-8")
        encoded_timestamp = timestamp.to_bytes(8, "big")
        encoded_nonce = nonce.encode("utf-8")

        return b"".join(
            [
                int(AuthMessageType.CHECK_SUBSCRIPTIONS).to_bytes(1, "big"),
                encoded_timestamp,
                len(encoded_nonce).to_bytes(1, "big"),
                encoded_nonce,
                len(tracing).to_bytes(2, "big"),
                tracing,
                len(encoded_url).to_bytes(2, "big"),
                encoded_url,
            ]
        )

    async def authorize_check_subscriptions(
        self, /, *, tracing: bytes, url: str, now: float
    ) -> Optional[str]:
        nonce = make_nonce()
        to_sign = self._prepare_check_subscriptions(
            tracing=tracing, url=url, timestamp=int(now), nonce=nonce
        )
        return sign(secret=self.secret, to_sign=to_sign, nonce=nonce, now=now)

    async def is_check_subscriptions_allowed(
        self, /, *, tracing: bytes, url: str, now: float, authorization: Optional[str]
    ) -> AuthResult:
        token = get_token(authorization, now=now, token_lifetime=self.token_lifetime)
        if token.type == TokenInfoType.UNAUTHORIZED:
            return AuthResult.UNAUTHORIZED
        if token.type == TokenInfoType.FORBIDDEN:
            return AuthResult.FORBIDDEN

        to_sign = self._prepare_check_subscriptions(
            tracing=tracing, url=url, timestamp=token.timestamp, nonce=token.nonce
        )
        return await check_code(
            secret=self.secret, to_sign=to_sign, code=token.hmac, db=self.db_config
        )

    def _prepare_set_subscriptions(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        strong_etag: StrongEtag,
        timestamp: int,
        nonce: str,
    ) -> bytes:
        encoded_url = url.encode("utf-8")
        encoded_etag = strong_etag.format.to_bytes(1, "big") + strong_etag.etag
        encoded_timestamp = timestamp.to_bytes(8, "big")
        encoded_nonce = nonce.encode("utf-8")

        return b"".join(
            [
                int(AuthMessageType.SET_SUBSCRIPTIONS).to_bytes(1, "big"),
                encoded_timestamp,
                len(encoded_nonce).to_bytes(1, "big"),
                encoded_nonce,
                len(tracing).to_bytes(2, "big"),
                tracing,
                len(encoded_url).to_bytes(2, "big"),
                encoded_url,
                encoded_etag,
            ]
        )

    async def authorize_set_subscriptions(
        self, /, *, tracing: bytes, url: str, strong_etag: StrongEtag, now: float
    ) -> Optional[str]:
        nonce = make_nonce()
        to_sign = self._prepare_set_subscriptions(
            tracing=tracing,
            url=url,
            strong_etag=strong_etag,
            timestamp=int(now),
            nonce=nonce,
        )
        return sign(secret=self.secret, to_sign=to_sign, nonce=nonce, now=now)

    async def is_set_subscriptions_allowed(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        strong_etag: StrongEtag,
        subscriptions: SetSubscriptionsInfo,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        token = get_token(authorization, now=now, token_lifetime=self.token_lifetime)
        if token.type == TokenInfoType.UNAUTHORIZED:
            return AuthResult.UNAUTHORIZED
        if token.type == TokenInfoType.FORBIDDEN:
            return AuthResult.FORBIDDEN

        to_sign = self._prepare_set_subscriptions(
            tracing=tracing,
            url=url,
            strong_etag=strong_etag,
            timestamp=token.timestamp,
            nonce=token.nonce,
        )
        return await check_code(
            secret=self.secret, to_sign=to_sign, code=token.hmac, db=self.db_config
        )

    def _prepare_stateful_continue_receive(
        self,
        /,
        *,
        tracing: bytes,
        identifier: bytes,
        part_id: int,
        url: str,
        timestamp: int,
        nonce: str,
    ) -> bytes:
        encoded_url = url.encode("utf-8")
        encoded_timestamp = timestamp.to_bytes(8, "big")
        encoded_nonce = nonce.encode("utf-8")

        return b"".join(
            [
                int(AuthMessageType.STATEFUL_CONTINUE_RECEIVE).to_bytes(1, "big"),
                encoded_timestamp,
                len(encoded_nonce).to_bytes(1, "big"),
                encoded_nonce,
                len(tracing).to_bytes(2, "big"),
                tracing,
                len(identifier).to_bytes(1, "big"),
                identifier,
                part_id.to_bytes(1, "big"),
                len(encoded_url).to_bytes(2, "big"),
                encoded_url,
            ]
        )

    async def authorize_stateful_continue_receive(
        self,
        /,
        *,
        tracing: bytes,
        identifier: bytes,
        part_id: int,
        url: str,
        now: float,
    ) -> Optional[str]:
        nonce = make_nonce()
        to_sign = self._prepare_stateful_continue_receive(
            tracing=tracing,
            identifier=identifier,
            part_id=part_id,
            url=url,
            timestamp=int(now),
            nonce=nonce,
        )
        return sign(secret=self.secret, to_sign=to_sign, nonce=nonce, now=now)

    async def is_stateful_continue_receive_allowed(
        self, /, *, url: str, message: S2B_ContinueReceive, now: float
    ) -> AuthResult:
        token = get_token(
            message.authorization, now=now, token_lifetime=self.token_lifetime
        )
        if token.type == TokenInfoType.UNAUTHORIZED:
            return AuthResult.UNAUTHORIZED
        if token.type == TokenInfoType.FORBIDDEN:
            return AuthResult.FORBIDDEN

        to_sign = self._prepare_stateful_continue_receive(
            tracing=message.tracing,
            identifier=message.identifier,
            part_id=message.part_id,
            url=url,
            timestamp=token.timestamp,
            nonce=token.nonce,
        )
        return await check_code(
            secret=self.secret, to_sign=to_sign, code=token.hmac, db=self.db_config
        )

    def _prepare_confirm_receive(
        self,
        /,
        *,
        tracing: bytes,
        identifier: bytes,
        num_subscribers: int,
        url: str,
        timestamp: int,
        nonce: str,
    ) -> bytes:
        encoded_url = url.encode("utf-8")
        encoded_timestamp = timestamp.to_bytes(8, "big")
        encoded_nonce = nonce.encode("utf-8")

        return b"".join(
            [
                int(AuthMessageType.CONFIRM_RECEIVE).to_bytes(1, "big"),
                encoded_timestamp,
                len(encoded_nonce).to_bytes(1, "big"),
                encoded_nonce,
                len(tracing).to_bytes(2, "big"),
                tracing,
                len(identifier).to_bytes(1, "big"),
                identifier,
                num_subscribers.to_bytes(4, "big"),
                len(encoded_url).to_bytes(2, "big"),
                encoded_url,
            ]
        )

    async def authorize_confirm_receive(
        self,
        /,
        *,
        tracing: bytes,
        identifier: bytes,
        num_subscribers: int,
        url: str,
        now: float,
    ) -> Optional[str]:
        nonce = make_nonce()
        to_sign = self._prepare_confirm_receive(
            tracing=tracing,
            identifier=identifier,
            num_subscribers=num_subscribers,
            url=url,
            timestamp=int(now),
            nonce=nonce,
        )
        return sign(secret=self.secret, to_sign=to_sign, nonce=nonce, now=now)

    async def is_confirm_receive_allowed(
        self,
        /,
        *,
        tracing: bytes,
        identifier: bytes,
        num_subscribers: int,
        url: str,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        token = get_token(authorization, now=now, token_lifetime=self.token_lifetime)
        if token.type == TokenInfoType.UNAUTHORIZED:
            return AuthResult.UNAUTHORIZED
        if token.type == TokenInfoType.FORBIDDEN:
            return AuthResult.FORBIDDEN

        to_sign = self._prepare_confirm_receive(
            tracing=tracing,
            identifier=identifier,
            num_subscribers=num_subscribers,
            url=url,
            timestamp=token.timestamp,
            nonce=token.nonce,
        )
        return await check_code(
            secret=self.secret, to_sign=to_sign, code=token.hmac, db=self.db_config
        )

    async def _prepare_confirm_missed(
        self, /, *, tracing: bytes, topic: bytes, url: str, timestamp: int, nonce: str
    ) -> bytes:
        encoded_url = url.encode("utf-8")
        encoded_timestamp = timestamp.to_bytes(8, "big")
        encoded_nonce = nonce.encode("utf-8")

        return b"".join(
            [
                int(AuthMessageType.CONFIRM_MISSED).to_bytes(1, "big"),
                encoded_timestamp,
                len(encoded_nonce).to_bytes(1, "big"),
                encoded_nonce,
                len(tracing).to_bytes(2, "big"),
                tracing,
                len(topic).to_bytes(2, "big"),
                topic,
                len(encoded_url).to_bytes(2, "big"),
                encoded_url,
            ]
        )

    async def authorize_confirm_missed(
        self, /, *, tracing: bytes, topic: bytes, url: str, now: float
    ) -> Optional[str]:
        nonce = make_nonce()
        to_sign = await self._prepare_confirm_missed(
            tracing=tracing, topic=topic, url=url, timestamp=int(now), nonce=nonce
        )
        return sign(secret=self.secret, to_sign=to_sign, nonce=nonce, now=now)

    async def is_confirm_missed_allowed(
        self,
        /,
        *,
        tracing: bytes,
        topic: bytes,
        url: str,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        token = get_token(authorization, now=now, token_lifetime=self.token_lifetime)
        if token.type == TokenInfoType.UNAUTHORIZED:
            return AuthResult.UNAUTHORIZED
        if token.type == TokenInfoType.FORBIDDEN:
            return AuthResult.FORBIDDEN

        to_sign = await self._prepare_confirm_missed(
            tracing=tracing,
            topic=topic,
            url=url,
            timestamp=token.timestamp,
            nonce=token.nonce,
        )
        return await check_code(
            secret=self.secret, to_sign=to_sign, code=token.hmac, db=self.db_config
        )


class ToSubscriberHmacAuth:
    """Sets up and allows requests to the subscriber which include hash-based
    message authentication codes. These tokens can only be generated if the
    sender knows the shared secret.

    Specifically, the authorization header is formatted as follows:
    `X-HMAC <timestamp>:<nonce>:<token>`, where timestamp is integer seconds from the epoch,
    and the token incorporates the relevant information and is signed with the
    shared secret.
    """

    def __init__(
        self,
        *,
        secret: str,
        token_lifetime: float = 120,
        db_config: IncomingHmacAuthDBConfig,
    ) -> None:
        self.secret = base64.urlsafe_b64decode(secret + "==")
        """The shared secret used to generate the HMAC tokens"""
        assert len(self.secret) == 64, "secret must be 64 bytes long"

        self.token_lifetime = token_lifetime
        """How long after a token is created that we still accept it"""

        self.db_config = db_config
        """The configuration for the recent codes database"""

    async def setup_to_subscriber_auth(self) -> None:
        await self.db_config.setup_hmac_auth_db()

    async def teardown_to_subscriber_auth(self) -> None:
        await self.db_config.teardown_hmac_auth_db()

    def _prepare_receive(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        topic: bytes,
        message_sha512: bytes,
        identifier: bytes,
        timestamp: int,
        nonce: str,
    ) -> bytes:
        assert len(message_sha512) == 64, "message_sha512 must be 64 bytes long"
        encoded_url = url.encode("utf-8")
        encoded_timestamp = timestamp.to_bytes(8, "big")
        encoded_nonce = nonce.encode("utf-8")

        return b"".join(
            [
                int(AuthMessageType.RECEIVE).to_bytes(1, "big"),
                encoded_timestamp,
                len(encoded_nonce).to_bytes(1, "big"),
                encoded_nonce,
                len(tracing).to_bytes(2, "big"),
                tracing,
                len(encoded_url).to_bytes(2, "big"),
                encoded_url,
                len(topic).to_bytes(2, "big"),
                topic,
                len(identifier).to_bytes(1, "big"),
                identifier,
                message_sha512,
            ]
        )

    async def authorize_receive(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        topic: bytes,
        message_sha512: bytes,
        identifier: bytes,
        now: float,
    ) -> Optional[str]:
        assert len(message_sha512) == 64, "message_sha512 must be 64 bytes long"
        nonce = make_nonce()
        to_sign = self._prepare_receive(
            tracing=tracing,
            url=url,
            topic=topic,
            message_sha512=message_sha512,
            identifier=identifier,
            timestamp=int(now),
            nonce=nonce,
        )
        return sign(secret=self.secret, to_sign=to_sign, nonce=nonce, now=now)

    async def is_receive_allowed(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        topic: bytes,
        message_sha512: bytes,
        identifier: bytes,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        assert len(message_sha512) == 64, "message_sha512 must be 64 bytes long"
        token = get_token(authorization, now=now, token_lifetime=self.token_lifetime)
        if token.type == TokenInfoType.UNAUTHORIZED:
            return AuthResult.UNAUTHORIZED
        if token.type == TokenInfoType.FORBIDDEN:
            return AuthResult.FORBIDDEN

        to_sign = self._prepare_receive(
            tracing=tracing,
            url=url,
            topic=topic,
            message_sha512=message_sha512,
            identifier=identifier,
            timestamp=token.timestamp,
            nonce=token.nonce,
        )
        return await check_code(
            secret=self.secret, to_sign=to_sign, code=token.hmac, db=self.db_config
        )

    def _prepare_missed(
        self,
        /,
        *,
        tracing: bytes,
        recovery: str,
        topic: bytes,
        timestamp: int,
        nonce: str,
    ) -> bytes:
        encoded_recovery = recovery.encode("utf-8")
        encoded_timestamp = timestamp.to_bytes(8, "big")
        encoded_nonce = nonce.encode("utf-8")

        return b"".join(
            [
                int(AuthMessageType.MISSED).to_bytes(1, "big"),
                encoded_timestamp,
                len(encoded_nonce).to_bytes(1, "big"),
                encoded_nonce,
                len(tracing).to_bytes(2, "big"),
                tracing,
                len(encoded_recovery).to_bytes(2, "big"),
                encoded_recovery,
                len(topic).to_bytes(2, "big"),
                topic,
            ]
        )

    async def authorize_missed(
        self, /, *, tracing: bytes, recovery: str, topic: bytes, now: float
    ) -> Optional[str]:
        nonce = make_nonce()
        to_sign = self._prepare_missed(
            tracing=tracing,
            recovery=recovery,
            topic=topic,
            timestamp=int(now),
            nonce=nonce,
        )
        return sign(secret=self.secret, to_sign=to_sign, nonce=nonce, now=now)

    async def is_missed_allowed(
        self,
        /,
        *,
        tracing: bytes,
        recovery: str,
        topic: bytes,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        token = get_token(authorization, now=now, token_lifetime=self.token_lifetime)
        if token.type == TokenInfoType.UNAUTHORIZED:
            return AuthResult.UNAUTHORIZED
        if token.type == TokenInfoType.FORBIDDEN:
            return AuthResult.FORBIDDEN
        to_sign = self._prepare_missed(
            tracing=tracing,
            recovery=recovery,
            topic=topic,
            timestamp=token.timestamp,
            nonce=token.nonce,
        )
        return await check_code(
            secret=self.secret, to_sign=to_sign, code=token.hmac, db=self.db_config
        )

    def _prepare_confirm_subscribe_exact(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        exact: bytes,
        timestamp: int,
        nonce: str,
    ) -> bytes:
        encoded_url = url.encode("utf-8")
        encoded_recovery = b"" if recovery is None else recovery.encode("utf-8")
        encoded_timestamp = timestamp.to_bytes(8, "big")
        encoded_nonce = nonce.encode("utf-8")

        return b"".join(
            [
                int(AuthMessageType.CONFIRM_SUBSCRIBE_EXACT).to_bytes(1, "big"),
                encoded_timestamp,
                len(encoded_nonce).to_bytes(1, "big"),
                encoded_nonce,
                len(tracing).to_bytes(2, "big"),
                tracing,
                len(encoded_url).to_bytes(2, "big"),
                encoded_url,
                len(encoded_recovery).to_bytes(2, "big"),
                encoded_recovery,
                len(exact).to_bytes(2, "big"),
                exact,
            ]
        )

    async def authorize_confirm_subscribe_exact(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        exact: bytes,
        now: float,
    ) -> Optional[str]:
        nonce = make_nonce()
        to_sign = self._prepare_confirm_subscribe_exact(
            tracing=tracing,
            url=url,
            recovery=recovery,
            exact=exact,
            timestamp=int(now),
            nonce=nonce,
        )
        return sign(secret=self.secret, to_sign=to_sign, nonce=nonce, now=now)

    async def is_confirm_subscribe_exact_allowed(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        exact: bytes,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        token = get_token(authorization, now=now, token_lifetime=self.token_lifetime)
        if token.type == TokenInfoType.UNAUTHORIZED:
            return AuthResult.UNAUTHORIZED
        if token.type == TokenInfoType.FORBIDDEN:
            return AuthResult.FORBIDDEN

        to_sign = self._prepare_confirm_subscribe_exact(
            tracing=tracing,
            url=url,
            recovery=recovery,
            exact=exact,
            timestamp=token.timestamp,
            nonce=token.nonce,
        )
        return await check_code(
            secret=self.secret, to_sign=to_sign, code=token.hmac, db=self.db_config
        )

    def _prepare_confirm_subscribe_glob(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        glob: str,
        timestamp: int,
        nonce: str,
    ) -> bytes:
        encoded_url = url.encode("utf-8")
        encoded_recovery = b"" if recovery is None else recovery.encode("utf-8")
        encoded_timestamp = timestamp.to_bytes(8, "big")
        encoded_glob = glob.encode("utf-8")
        encoded_nonce = nonce.encode("utf-8")

        return b"".join(
            [
                int(AuthMessageType.CONFIRM_SUBSCRIBE_GLOB).to_bytes(1, "big"),
                encoded_timestamp,
                len(encoded_nonce).to_bytes(1, "big"),
                encoded_nonce,
                len(tracing).to_bytes(2, "big"),
                tracing,
                len(encoded_url).to_bytes(2, "big"),
                encoded_url,
                len(encoded_recovery).to_bytes(2, "big"),
                encoded_recovery,
                len(encoded_glob).to_bytes(2, "big"),
                encoded_glob,
            ]
        )

    async def authorize_confirm_subscribe_glob(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        glob: str,
        now: float,
    ) -> Optional[str]:
        nonce = make_nonce()
        to_sign = self._prepare_confirm_subscribe_glob(
            tracing=tracing,
            url=url,
            recovery=recovery,
            glob=glob,
            timestamp=int(now),
            nonce=nonce,
        )
        return sign(secret=self.secret, to_sign=to_sign, nonce=nonce, now=now)

    async def is_confirm_subscribe_glob_allowed(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        glob: str,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        token = get_token(authorization, now=now, token_lifetime=self.token_lifetime)
        if token.type == TokenInfoType.UNAUTHORIZED:
            return AuthResult.UNAUTHORIZED
        if token.type == TokenInfoType.FORBIDDEN:
            return AuthResult.FORBIDDEN

        to_sign = self._prepare_confirm_subscribe_glob(
            tracing=tracing,
            url=url,
            recovery=recovery,
            glob=glob,
            timestamp=token.timestamp,
            nonce=token.nonce,
        )
        return await check_code(
            secret=self.secret, to_sign=to_sign, code=token.hmac, db=self.db_config
        )

    async def _prepare_confirm_notify(
        self,
        /,
        *,
        tracing: bytes,
        identifier: bytes,
        subscribers: int,
        topic: bytes,
        message_sha512: bytes,
        timestamp: int,
        nonce: str,
    ) -> bytes:
        assert len(message_sha512) == 64, "message_sha512 must be 64 bytes long"
        encoded_timestamp = timestamp.to_bytes(8, "big")
        encoded_nonce = nonce.encode("utf-8")

        return b"".join(
            [
                int(AuthMessageType.CONFIRM_NOTIFY).to_bytes(1, "big"),
                encoded_timestamp,
                len(encoded_nonce).to_bytes(1, "big"),
                encoded_nonce,
                len(tracing).to_bytes(2, "big"),
                tracing,
                len(identifier).to_bytes(1, "big"),
                identifier,
                subscribers.to_bytes(4, "big"),
                len(topic).to_bytes(2, "big"),
                topic,
                message_sha512,
            ]
        )

    async def authorize_confirm_notify(
        self,
        /,
        *,
        tracing: bytes,
        identifier: bytes,
        subscribers: int,
        topic: bytes,
        message_sha512: bytes,
        now: float,
    ) -> Optional[str]:
        assert len(message_sha512) == 64, "message_sha512 must be 64 bytes long"
        nonce = make_nonce()
        to_sign = await self._prepare_confirm_notify(
            tracing=tracing,
            identifier=identifier,
            subscribers=subscribers,
            topic=topic,
            message_sha512=message_sha512,
            timestamp=int(now),
            nonce=nonce,
        )
        return sign(secret=self.secret, to_sign=to_sign, nonce=nonce, now=now)

    async def is_confirm_notify_allowed(
        self,
        /,
        *,
        tracing: bytes,
        identifier: bytes,
        subscribers: int,
        topic: bytes,
        message_sha512: bytes,
        authorization: Optional[str],
        now: float,
    ) -> AuthResult:
        assert len(message_sha512) == 64, "message_sha512 must be 64 bytes long"
        token = get_token(authorization, now=now, token_lifetime=self.token_lifetime)
        if token.type == TokenInfoType.UNAUTHORIZED:
            return AuthResult.UNAUTHORIZED
        if token.type == TokenInfoType.FORBIDDEN:
            return AuthResult.FORBIDDEN

        to_sign = await self._prepare_confirm_notify(
            tracing=tracing,
            identifier=identifier,
            subscribers=subscribers,
            topic=topic,
            message_sha512=message_sha512,
            timestamp=token.timestamp,
            nonce=token.nonce,
        )
        return await check_code(
            secret=self.secret, to_sign=to_sign, code=token.hmac, db=self.db_config
        )

    def _prepare_stateful_confirm_configure(
        self, /, *, broadcaster_nonce: bytes, tracing: bytes, timestamp: int, nonce: str
    ) -> bytes:
        encoded_timestamp = timestamp.to_bytes(8, "big")
        encoded_nonce = nonce.encode("utf-8")

        return b"".join(
            [
                int(AuthMessageType.STATEFUL_CONFIRM_CONFIGURE).to_bytes(1, "big"),
                encoded_timestamp,
                len(encoded_nonce).to_bytes(1, "big"),
                encoded_nonce,
                len(tracing).to_bytes(2, "big"),
                tracing,
                len(broadcaster_nonce).to_bytes(1, "big"),
                broadcaster_nonce,
            ]
        )

    async def authorize_stateful_confirm_configure(
        self, /, *, broadcaster_nonce: bytes, tracing: bytes, now: float
    ) -> Optional[str]:
        nonce = make_nonce()
        to_sign = self._prepare_stateful_confirm_configure(
            broadcaster_nonce=broadcaster_nonce,
            tracing=tracing,
            timestamp=int(now),
            nonce=nonce,
        )
        return sign(secret=self.secret, to_sign=to_sign, nonce=nonce, now=now)

    async def is_stateful_confirm_configure_allowed(
        self, /, *, message: B2S_ConfirmConfigure, now: float
    ) -> AuthResult:
        token = get_token(
            message.authorization, now=now, token_lifetime=self.token_lifetime
        )
        if token.type == TokenInfoType.UNAUTHORIZED:
            return AuthResult.UNAUTHORIZED
        if token.type == TokenInfoType.FORBIDDEN:
            return AuthResult.FORBIDDEN

        to_sign = self._prepare_stateful_confirm_configure(
            broadcaster_nonce=message.broadcaster_nonce,
            tracing=message.tracing,
            timestamp=token.timestamp,
            nonce=token.nonce,
        )
        return await check_code(
            secret=self.secret, to_sign=to_sign, code=token.hmac, db=self.db_config
        )

    async def _prepare_stateful_enable_zstd_preset(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        compressor_identifier: int,
        compression_level: int,
        min_size: int,
        max_size: int,
        timestamp: int,
        nonce: str,
    ) -> bytes:
        encoded_url = url.encode("utf-8")
        encoded_timestamp = timestamp.to_bytes(8, "big")
        encoded_nonce = nonce.encode("utf-8")

        return b"".join(
            [
                int(AuthMessageType.STATEFUL_ENABLE_ZSTD_PRESET).to_bytes(1, "big"),
                encoded_timestamp,
                len(encoded_nonce).to_bytes(1, "big"),
                encoded_nonce,
                len(tracing).to_bytes(2, "big"),
                tracing,
                len(encoded_url).to_bytes(2, "big"),
                encoded_url,
                compressor_identifier.to_bytes(4, "big"),
                compression_level.to_bytes(4, "big"),
                min_size.to_bytes(8, "big"),
                max_size.to_bytes(8, "big"),
            ]
        )

    async def authorize_stateful_enable_zstd_preset(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        compressor_identifier: int,
        compression_level: int,
        min_size: int,
        max_size: int,
        authorization: Optional[str],
        now: float,
    ) -> Optional[str]:
        nonce = make_nonce()
        to_sign = await self._prepare_stateful_enable_zstd_preset(
            tracing=tracing,
            url=url,
            compressor_identifier=compressor_identifier,
            compression_level=compression_level,
            min_size=min_size,
            max_size=max_size,
            timestamp=int(now),
            nonce=nonce,
        )
        return sign(secret=self.secret, to_sign=to_sign, nonce=nonce, now=now)

    async def is_stateful_enable_zstd_preset_allowed(
        self, /, *, url: str, message: B2S_EnableZstdPreset, now: float
    ) -> AuthResult:
        token = get_token(
            message.authorization, now=now, token_lifetime=self.token_lifetime
        )
        if token.type == TokenInfoType.UNAUTHORIZED:
            return AuthResult.UNAUTHORIZED
        if token.type == TokenInfoType.FORBIDDEN:
            return AuthResult.FORBIDDEN

        to_sign = await self._prepare_stateful_enable_zstd_preset(
            tracing=message.tracing,
            url=url,
            compressor_identifier=message.identifier,
            compression_level=message.compression_level,
            min_size=message.min_size,
            max_size=message.max_size,
            timestamp=token.timestamp,
            nonce=token.nonce,
        )
        return await check_code(
            secret=self.secret, to_sign=to_sign, code=token.hmac, db=self.db_config
        )

    async def _prepare_stateful_enable_zstd_custom(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        compressor_identifier: int,
        compression_level: int,
        min_size: int,
        max_size: int,
        sha512: bytes,
        timestamp: int,
        nonce: str,
    ) -> bytes:
        assert len(sha512) == 64, "sha512 must be 64 bytes long"
        encoded_url = url.encode("utf-8")
        encoded_timestamp = timestamp.to_bytes(8, "big")
        encoded_nonce = nonce.encode("utf-8")

        return b"".join(
            [
                int(AuthMessageType.STATEFUL_ENABLE_ZSTD_CUSTOM).to_bytes(1, "big"),
                encoded_timestamp,
                len(encoded_nonce).to_bytes(1, "big"),
                encoded_nonce,
                len(tracing).to_bytes(2, "big"),
                tracing,
                len(encoded_url).to_bytes(2, "big"),
                encoded_url,
                compressor_identifier.to_bytes(4, "big"),
                compression_level.to_bytes(4, "big"),
                min_size.to_bytes(8, "big"),
                max_size.to_bytes(8, "big"),
                sha512,
            ]
        )

    async def authorize_stateful_enable_zstd_custom(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        compressor_identifier: int,
        compression_level: int,
        min_size: int,
        max_size: int,
        sha512: bytes,
        now: float,
    ) -> Optional[str]:
        assert len(sha512) == 64, "sha512 must be 64 bytes long"
        nonce = make_nonce()
        to_sign = await self._prepare_stateful_enable_zstd_custom(
            tracing=tracing,
            url=url,
            compressor_identifier=compressor_identifier,
            compression_level=compression_level,
            min_size=min_size,
            max_size=max_size,
            sha512=sha512,
            timestamp=int(now),
            nonce=nonce,
        )
        return sign(secret=self.secret, to_sign=to_sign, nonce=nonce, now=now)

    async def is_stateful_enable_zstd_custom_allowed(
        self, /, *, url: str, message: B2S_EnableZstdCustom, now: float
    ) -> AuthResult:
        token = get_token(
            message.authorization, now=now, token_lifetime=self.token_lifetime
        )
        if token.type == TokenInfoType.UNAUTHORIZED:
            return AuthResult.UNAUTHORIZED
        if token.type == TokenInfoType.FORBIDDEN:
            return AuthResult.FORBIDDEN

        to_sign = await self._prepare_stateful_enable_zstd_custom(
            tracing=message.tracing,
            url=url,
            compressor_identifier=message.identifier,
            compression_level=message.compression_level,
            min_size=message.min_size,
            max_size=message.max_size,
            sha512=message.sha512,
            timestamp=token.timestamp,
            nonce=token.nonce,
        )
        return await check_code(
            secret=self.secret, to_sign=to_sign, code=token.hmac, db=self.db_config
        )

    def _prepare_stateful_disable_zstd_custom(
        self,
        /,
        *,
        tracing: bytes,
        compressor_identifier: int,
        url: str,
        timestamp: int,
        nonce: str,
    ) -> bytes:
        encoded_url = url.encode("utf-8")
        encoded_timestamp = timestamp.to_bytes(8, "big")
        encoded_nonce = nonce.encode("utf-8")

        return b"".join(
            [
                int(AuthMessageType.STATEFUL_DISABLE_ZSTD_CUSTOM).to_bytes(1, "big"),
                encoded_timestamp,
                len(encoded_nonce).to_bytes(1, "big"),
                encoded_nonce,
                len(tracing).to_bytes(2, "big"),
                tracing,
                compressor_identifier.to_bytes(4, "big"),
                len(encoded_url).to_bytes(2, "big"),
                encoded_url,
            ]
        )

    async def authorize_stateful_disable_zstd_custom(
        self, /, *, tracing: bytes, compressor_identifier: int, url: str, now: float
    ) -> Optional[str]:
        nonce = make_nonce()
        to_sign = self._prepare_stateful_disable_zstd_custom(
            tracing=tracing,
            compressor_identifier=compressor_identifier,
            url=url,
            timestamp=int(now),
            nonce=nonce,
        )
        return sign(secret=self.secret, to_sign=to_sign, nonce=nonce, now=now)

    async def is_stateful_disable_zstd_custom_allowed(
        self, /, *, url: str, message: B2S_DisableZstdCustom, now: float
    ) -> AuthResult:
        token = get_token(
            message.authorization, now=now, token_lifetime=self.token_lifetime
        )
        if token.type == TokenInfoType.UNAUTHORIZED:
            return AuthResult.UNAUTHORIZED
        if token.type == TokenInfoType.FORBIDDEN:
            return AuthResult.FORBIDDEN

        to_sign = self._prepare_stateful_disable_zstd_custom(
            tracing=message.tracing,
            compressor_identifier=message.identifier,
            url=url,
            timestamp=token.timestamp,
            nonce=token.nonce,
        )
        return await check_code(
            secret=self.secret, to_sign=to_sign, code=token.hmac, db=self.db_config
        )


if TYPE_CHECKING:
    __: Type[ToBroadcasterAuthConfig] = ToBroadcasterHmacAuth
    ___: Type[ToSubscriberAuthConfig] = ToSubscriberHmacAuth



from typing import Union

from lonelypsp.stateful.messages.configure import S2B_Configure
from lonelypsp.stateful.messages.confirm_configure import B2S_ConfirmConfigure
from lonelypsp.stateful.messages.confirm_notify import B2S_ConfirmNotify
from lonelypsp.stateful.messages.confirm_receive import S2B_ConfirmReceive
from lonelypsp.stateful.messages.confirm_subscribe import (
    B2S_ConfirmSubscribeExact,
    B2S_ConfirmSubscribeGlob,
)
from lonelypsp.stateful.messages.confirm_unsubscribe import (
    B2S_ConfirmUnsubscribeExact,
    B2S_ConfirmUnsubscribeGlob,
)
from lonelypsp.stateful.messages.continue_notify import B2S_ContinueNotify
from lonelypsp.stateful.messages.continue_receive import S2B_ContinueReceive
from lonelypsp.stateful.messages.disable_zstd_custom import B2S_DisableZstdCustom
from lonelypsp.stateful.messages.enable_zstd_custom import B2S_EnableZstdCustom
from lonelypsp.stateful.messages.enable_zstd_preset import B2S_EnableZstdPreset
from lonelypsp.stateful.messages.missed import B2S_Missed
from lonelypsp.stateful.messages.notify import S2B_Notify
from lonelypsp.stateful.messages.notify_stream import S2B_NotifyStream
from lonelypsp.stateful.messages.receive_stream import B2S_ReceiveStream
from lonelypsp.stateful.messages.subscribe import (
    S2B_SubscribeExact,
    S2B_SubscribeGlob,
)
from lonelypsp.stateful.messages.unsubscribe import (
    S2B_UnsubscribeExact,
    S2B_UnsubscribeGlob,
)

S2B_Message = Union[
    S2B_Configure,
    S2B_ConfirmReceive,
    S2B_ContinueReceive,
    S2B_NotifyStream,
    S2B_Notify,
    S2B_SubscribeExact,
    S2B_SubscribeGlob,
    S2B_UnsubscribeExact,
    S2B_UnsubscribeGlob,
]
"""Type alias for any message from a subscriber to a broadcaster"""

B2S_Message = Union[
    B2S_ConfirmConfigure,
    B2S_ConfirmNotify,
    B2S_ConfirmSubscribeExact,
    B2S_ConfirmSubscribeGlob,
    B2S_ConfirmUnsubscribeExact,
    B2S_ConfirmUnsubscribeGlob,
    B2S_ContinueNotify,
    B2S_DisableZstdCustom,
    B2S_EnableZstdCustom,
    B2S_EnableZstdPreset,
    B2S_Missed,
    B2S_ReceiveStream,
]
"""Type alias for any message from a broadcaster to a subscriber"""



import base64
import io
import json
import secrets
import struct
import time
import warnings
from types import TracebackType
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Type

from lonelypsp.auth.config import BadAuthResult
from lonelypsp.compat import fast_dataclass
from lonelypsp.tracing.impl.simple.config import SimpleTracingConfig
from lonelypsp.tracing.impl.simple.db import SimpleTracingDBSidecar
from lonelypsp.tracing.impl.simple.shared import cristians
from lonelypsp.tracing.shared.handle_trusted_notify import (
    HandledTrustedNotify,
    HandledTrustedNotifyHandleMissedStart,
    HandledTrustedNotifyReceivedResponse,
    HandledTrustedNotifySendingReceive,
    HandleTrustedNotifyHandleMissedDone,
    HandleTrustedNotifyResponseAuthResultReady,
    HandleTrustedNotifyUnsubscribeImmediateDone,
)
from lonelypsp.tracing.shared.tracing_and_followup import TracingAndFollowup
from lonelypsp.tracing.stateless.notify import (
    StatelessTracingNotifyOnAuthResult,
    StatelessTracingNotifyOnAuthVerified,
    StatelessTracingNotifyOnHashed,
    StatelessTracingNotifyOnReceivedFirst,
    StatelessTracingNotifyOnResponseAuthResult,
    StatelessTracingNotifyOnResponseReceived,
    StatelessTracingNotifyOnRetryDetermined,
    StatelessTracingNotifyOnSending,
    StatelessTracingNotifyOnSendingResponse,
    StatelessTracingNotifyStart,
    StatelessTracingNotifyStartFirst,
)


@fast_dataclass
class RemoteTrace:
    name: str
    extra_json: str
    occurred_at_remote: float


class SimpleStatelessTracingNotifyStart:
    """Subscriber side"""

    def __init__(self, db: SimpleTracingDBSidecar, config: SimpleTracingConfig) -> None:
        self.db: SimpleTracingDBSidecar = db
        """the database to use for the tracing data, assumed to already be entered"""

        self.config: SimpleTracingConfig = config
        """the configuration for the tracing data"""

        self.uid: bytes = int(time.time()).to_bytes(8, "big") + secrets.token_bytes(4)
        """the unique identifier we assigned without having to contact the db"""

        self.next_ord: int = 0
        """the order value for the next piece of timing information"""

        self.request_sent_at_local_nano: Optional[int] = None
        """when the request was sent in local nanotime"""

        self.response_received_at_local_nano: Optional[int] = None
        """when the response was received in local nanotime"""

        self._entered = False
        self._exited = False

    def __enter__(self) -> StatelessTracingNotifyStartFirst:
        assert not self._entered, "Cannot re-enter a tracing event"
        self._entered = True
        self.db.enqueue(
            [
                ("BEGIN IMMEDIATE TRANSACTION", []),
                (
                    "INSERT INTO stateless_notifies(uid, created_at) VALUES (?, ?)",
                    [self.uid, time.time()],
                ),
                ("COMMIT TRANSACTION", []),
            ]
        )
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        assert self._entered, "Cannot exit a tracing event that was never entered"
        if self._exited:
            return

        self._exited = True
        self.db.enqueue(
            [
                (
                    "BEGIN IMMEDIATE TRANSACTION",
                    [],
                ),
                (
                    "UPDATE stateless_notifies SET finished_at = ? WHERE uid = ?",
                    [time.time(), self.uid],
                ),
                (
                    "COMMIT TRANSACTION",
                    [],
                ),
            ]
        )

    def _reserve_ord(self, /) -> int:
        ord = self.next_ord
        self.next_ord += 1
        return ord

    def _trace(
        self,
        /,
        *,
        name: str,
        extra: Dict[str, Any],
        occurred_at: float,
        raw_occurred_at: Optional[float] = None,
    ) -> None:
        assert self._entered, "not entered"
        assert not self._exited, "already exited"
        extra_json = json.dumps(extra)
        if raw_occurred_at is None:
            raw_occurred_at = occurred_at
        self.db.enqueue(
            [
                (
                    "BEGIN IMMEDIATE TRANSACTION",
                    [],
                ),
                (
                    "INSERT INTO stateless_notify_timings "
                    "(notify_id, ord, name, extra, occurred_at, raw_occurred_at) "
                    "SELECT"
                    " stateless_notifies.id, ?, ?, ?, ? "
                    "FROM stateless_notifies "
                    "WHERE stateless_notifies.uid = ?",
                    [
                        self._reserve_ord(),
                        name,
                        extra_json,
                        occurred_at,
                        raw_occurred_at,
                        self.uid,
                    ],
                ),
                (
                    "COMMIT TRANSACTION",
                    [],
                ),
            ]
        )

    # StartFirst
    def _on_start(
        self, /, *, name: str, topic: bytes, length: int, filelike: bool
    ) -> None:
        assert self._entered, "not entered"
        assert not self._exited, "already exited"
        self.db.enqueue(
            [
                ("BEGIN IMMEDIATE TRANSACTION", []),
                (
                    "UPDATE stateless_notifies SET topic = ?, length = ? WHERE uid = ?",
                    [topic, length, self.uid],
                ),
                (
                    "INSERT INTO stateless_notify_timings "
                    "(notify_id, ord, name, extra, occurred_at) "
                    "SELECT"
                    " stateless_notifies.id, ?, ?, ?, ? "
                    "FROM stateless_notifies "
                    "WHERE stateless_notifies.uid = ?",
                    [
                        self._reserve_ord(),
                        name,
                        b'{"filelike": ' + (b"true}" if filelike else b"false}"),
                        time.time(),
                        self.uid,
                    ],
                ),
                ("COMMIT TRANSACTION", []),
            ]
        )

    def on_start_without_hash(
        self, /, *, topic: bytes, length: int, filelike: bool
    ) -> StatelessTracingNotifyOnHashed:
        self._on_start(
            name="start_without_hash",
            topic=topic,
            length=length,
            filelike=filelike,
        )
        return self

    def on_start_with_hash(
        self, /, *, topic: bytes, length: int, filelike: bool
    ) -> StatelessTracingNotifyOnSending:
        self._on_start(
            name="start_with_hash",
            topic=topic,
            length=length,
            filelike=filelike,
        )
        return self

    # OnHashed
    def on_hashed(self) -> StatelessTracingNotifyOnSending:
        self._trace(name="hashed", extra={}, occurred_at=time.time())
        return self

    # OnSending
    def on_sending_request(
        self, /, *, broadcaster: str, identifier: bytes
    ) -> TracingAndFollowup[StatelessTracingNotifyOnResponseReceived]:
        self._trace(
            name="sending_request",
            extra={"broadcaster": broadcaster},
            occurred_at=time.time(),
        )
        self.request_sent_at_local_nano = time.time_ns()
        return TracingAndFollowup(tracing=b"", followup=self)

    def on_network_error(self) -> StatelessTracingNotifyOnRetryDetermined:
        self._trace(name="network_error", extra={}, occurred_at=time.time())
        return self

    def on_response_received(
        self, /, *, status_code: int
    ) -> StatelessTracingNotifyOnResponseAuthResult:
        self.response_received_at_local_nano = time.time_ns()
        self._trace(
            name="response_received",
            extra={"status_code": status_code},
            occurred_at=time.time(),
        )
        return self

    # StatelessTracingNotifyOnRetryDetermined
    def on_retry_prevented(self) -> None:
        self._trace(name="retry_prevented", extra={}, occurred_at=time.time())

    def on_retries_exhausted(self) -> None:
        self._trace(name="retries_exhausted", extra={}, occurred_at=time.time())

    def on_waiting_to_retry(self) -> StatelessTracingNotifyOnRetryDetermined:
        self._trace(name="waiting_to_retry", extra={}, occurred_at=time.time())
        return self

    def on_retrying(self) -> StatelessTracingNotifyOnSending:
        self._trace(name="retrying", extra={}, occurred_at=time.time())
        return self

    # StatelessTracingNotifyOnResponseAuthResult
    def on_bad_response(self) -> StatelessTracingNotifyOnRetryDetermined:
        self._trace(name="bad_response", extra={}, occurred_at=time.time())
        return self

    def on_bad_auth_result(
        self, /, *, result: BadAuthResult
    ) -> StatelessTracingNotifyOnRetryDetermined:
        self._trace(
            name="bad_auth_result",
            extra={"result": result.name},
            occurred_at=time.time(),
        )
        return self

    def on_response_notify_accepted(
        self, /, *, tracing: bytes, num_subscribers: int
    ) -> None:
        assert self._entered, "not entered"
        assert not self._exited, "already exited"
        assert self.request_sent_at_local_nano is not None, "no request sent"
        assert self.response_received_at_local_nano is not None, "no response received"
        occurred_at = time.time()

        rdr = io.BytesIO(tracing)
        truncated = rdr.read(1) != b"\0"
        num_traces = int.from_bytes(rdr.read(2), "big")
        traces: List[RemoteTrace] = []
        for _ in range(num_traces):
            name_len = int.from_bytes(rdr.read(2), "big")
            name = rdr.read(name_len).decode("utf-8")
            extra_len = int.from_bytes(rdr.read(2), "big")
            extra = rdr.read(extra_len).decode("utf-8")
            occurred_at_remote = struct.unpack(">d", rdr.read(8))[0]
            traces.append(
                RemoteTrace(
                    name=name,
                    extra_json=extra,
                    occurred_at_remote=occurred_at_remote,
                )
            )

        request_received_at_nano_remote = int.from_bytes(rdr.read(8), "big")
        response_sent_at_nano_remote = int.from_bytes(rdr.read(8), "big")
        offset = cristians.compute_offset(
            cristians.CristiansEnd(
                request_sent_at_local_nano=self.request_sent_at_local_nano,
                request_received_at_remote_nano=request_received_at_nano_remote,
                response_sent_at_remote_nano=response_sent_at_nano_remote,
                response_received_at_local_nano=self.response_received_at_local_nano,
            )
        )

        if truncated:
            warnings.warn("tracing data was truncated")

        to_enqueue: List[Tuple[str, List[Any]]] = [
            ("BEGIN IMMEDIATE TRANSACTION", []),
            (
                "INSERT INTO stateless_notify_timings "
                "(notify_id, ord, name, extra, occurred_at, raw_occurred_at) "
                "SELECT"
                " stateless_notifies.id, ?, ?, ?, ? "
                "FROM stateless_notifies "
                "WHERE stateless_notifies.uid = ?",
                [
                    self._reserve_ord(),
                    name,
                    json.dumps(
                        {
                            "num_subscribers": num_subscribers,
                            "cristians": {
                                "request_sent_at_local_nano": self.request_sent_at_local_nano,
                                "request_received_at_remote_nano": request_received_at_nano_remote,
                                "response_sent_at_remote_nano": response_sent_at_nano_remote,
                                "response_received_at_local_nano": self.response_received_at_local_nano,
                                "latency_ms": offset.latency,
                                "offset_ms": offset.offset,
                            },
                            "truncated": truncated,
                        }
                    ),
                    occurred_at,
                    occurred_at,
                    self.uid,
                ],
            ),
        ]

        for remote_trace in traces:
            to_enqueue.append(
                (
                    "INSERT INTO stateless_notify_timings "
                    "(notify_id, ord, name, extra, occurred_at, raw_occurred_at) "
                    "SELECT"
                    " stateless_notifies.id, ?, ?, ?, ? "
                    "FROM stateless_notifies "
                    "WHERE stateless_notifies.uid = ?",
                    [
                        self._reserve_ord(),
                        f"remote_{remote_trace.name}",
                        remote_trace.extra_json,
                        remote_trace.occurred_at_remote - offset.offset,
                        remote_trace.occurred_at_remote,
                    ],
                )
            )

        to_enqueue.append(("COMMIT TRANSACTION", []))
        self.db.enqueue(to_enqueue)


class SimpleStatelessTracingNotifyOnReceived:
    """Broadcaster side"""

    def __init__(self) -> None:
        self.request_received_at_nano: Optional[int] = None
        """when the request was received in nanoseconds"""

        self.traces: List[RemoteTrace] = []
        """the traces we will send back as a broadcaster"""

    def __enter__(self) -> StatelessTracingNotifyOnReceivedFirst:
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None: ...

    # OnReceivedFirst
    def on_received(self) -> StatelessTracingNotifyOnAuthResult:
        self.request_received_at_nano = time.time_ns()
        return self

    # OnAuthResult
    def on_bad_request(self) -> None: ...

    def on_bad_auth_result(self, /, *, result: BadAuthResult) -> None: ...

    def on_auth_tentatively_accepted(self) -> StatelessTracingNotifyOnAuthVerified:
        self.traces.append(
            RemoteTrace(
                name="auth_tentatively_accepted",
                extra_json="{}",
                occurred_at_remote=time.time(),
            )
        )
        return self

    # OnAuthVerified
    def on_auth_mismatch(self) -> None: ...

    def on_auth_accepted(
        self, topic: bytes, length: int, identifier: bytes, tracing: bytes
    ) -> HandledTrustedNotify[StatelessTracingNotifyOnSendingResponse]:
        self.traces.append(
            RemoteTrace(
                name="auth_accepted",
                extra_json="{}",
                occurred_at_remote=time.time(),
            )
        )
        return self

    # HandledTrustedNotify
    def on_unavailable(self) -> None: ...

    def on_exact_subscriber_found(
        self, /, *, url: str
    ) -> HandledTrustedNotifySendingReceive[
        HandledTrustedNotify[StatelessTracingNotifyOnSendingResponse]
    ]:
        self.traces.append(
            RemoteTrace(
                name="exact_subscriber_found",
                extra_json=json.dumps({"url": url}),
                occurred_at_remote=time.time(),
            )
        )
        return self

    def on_glob_subscriber_found(
        self, /, *, glob: str, url: str
    ) -> HandledTrustedNotifySendingReceive[
        HandledTrustedNotify[StatelessTracingNotifyOnSendingResponse]
    ]:
        self.traces.append(
            RemoteTrace(
                name="glob_subscriber_found",
                extra_json=json.dumps({"glob": glob, "url": url}),
                occurred_at_remote=time.time(),
            )
        )
        return self

    def on_no_more_subscribers(self) -> StatelessTracingNotifyOnSendingResponse:
        self.traces.append(
            RemoteTrace(
                name="no_more_subscribers",
                extra_json="{}",
                occurred_at_remote=time.time(),
            )
        )
        return self

    # HandledTrustedNotifySendingReceive
    def on_sending_receive(
        self, /, *, identifier: bytes
    ) -> TracingAndFollowup[
        HandledTrustedNotifyReceivedResponse[
            HandledTrustedNotify[StatelessTracingNotifyOnSendingResponse]
        ]
    ]:
        self.traces.append(
            RemoteTrace(
                name="sending_receive",
                extra_json=json.dumps(
                    {
                        "identifier_b64url": base64.urlsafe_b64encode(
                            identifier
                        ).decode("ascii")
                    }
                ),
                occurred_at_remote=time.time(),
            )
        )
        return TracingAndFollowup(tracing=b"", followup=self)

    # HandledTrustedNotifyReceivedResponse
    def on_network_error(
        self,
    ) -> HandledTrustedNotifyHandleMissedStart[
        HandledTrustedNotify[StatelessTracingNotifyOnSendingResponse]
    ]:
        self.traces.append(
            RemoteTrace(
                name="network_error",
                extra_json="{}",
                occurred_at_remote=time.time(),
            )
        )
        return self

    def on_response_received(
        self, /, *, status_code: int
    ) -> HandleTrustedNotifyResponseAuthResultReady[
        HandledTrustedNotify[StatelessTracingNotifyOnSendingResponse]
    ]:
        self.traces.append(
            RemoteTrace(
                name="response_received",
                extra_json=json.dumps({"status_code": status_code}),
                occurred_at_remote=time.time(),
            )
        )
        return self

    # HandledTrustedNotifyHandleMissedStart
    def on_handle_missed_start(
        self,
    ) -> HandleTrustedNotifyHandleMissedDone[
        HandledTrustedNotify[StatelessTracingNotifyOnSendingResponse]
    ]:
        self.traces.append(
            RemoteTrace(
                name="handle_missed_start",
                extra_json="{}",
                occurred_at_remote=time.time(),
            )
        )
        return self

    # HandleTrustedNotifyHandleMissedDone
    def on_handle_missed_skipped(
        self, /, *, recovery: Optional[str], next_retry_at: Optional[float]
    ) -> HandledTrustedNotify[StatelessTracingNotifyOnSendingResponse]:
        self.traces.append(
            RemoteTrace(
                name="handle_missed_skipped",
                extra_json=json.dumps(
                    {"recovery": recovery, "next_retry_at": next_retry_at}
                ),
                occurred_at_remote=time.time(),
            )
        )
        return self

    def on_handle_missed_success(
        self, recovery: str, next_retry_at: float
    ) -> HandledTrustedNotify[StatelessTracingNotifyOnSendingResponse]:
        self.traces.append(
            RemoteTrace(
                name="handle_missed_success",
                extra_json=json.dumps(
                    {"recovery": recovery, "next_retry_at": next_retry_at}
                ),
                occurred_at_remote=time.time(),
            )
        )
        return self

    def on_handle_missed_unavailable(
        self, recovery: str, next_retry_at: float
    ) -> HandledTrustedNotify[StatelessTracingNotifyOnSendingResponse]:
        self.traces.append(
            RemoteTrace(
                name="handle_missed_unavailable",
                extra_json=json.dumps(
                    {"recovery": recovery, "next_retry_at": next_retry_at}
                ),
                occurred_at_remote=time.time(),
            )
        )
        return self

    # HandleTrustedNotifyResponseAuthResultReady
    def on_bad_receive_response(
        self,
    ) -> HandledTrustedNotifyHandleMissedStart[
        HandledTrustedNotify[StatelessTracingNotifyOnSendingResponse]
    ]:
        self.traces.append(
            RemoteTrace(
                name="bad_response",
                extra_json="{}",
                occurred_at_remote=time.time(),
            )
        )
        return self

    def on_unsubscribe_immediate_requested(
        self,
    ) -> HandleTrustedNotifyUnsubscribeImmediateDone[
        HandledTrustedNotify[StatelessTracingNotifyOnSendingResponse]
    ]:
        self.traces.append(
            RemoteTrace(
                name="unsubscribe_immediate_requested",
                extra_json="{}",
                occurred_at_remote=time.time(),
            )
        )
        return self

    # HandleTrustedNotifyResponseAuthResultReady
    def on_bad_receive_auth_result(
        self, /, *, result: BadAuthResult
    ) -> HandledTrustedNotifyHandleMissedStart[
        HandledTrustedNotify[StatelessTracingNotifyOnSendingResponse]
    ]:
        self.traces.append(
            RemoteTrace(
                name="bad_auth_result",
                extra_json=json.dumps({"result": result.name}),
                occurred_at_remote=time.time(),
            )
        )
        return self

    def on_receive_confirmed(
        self, /, *, tracing: bytes, num_subscribers: int
    ) -> HandledTrustedNotify[StatelessTracingNotifyOnSendingResponse]:
        self.traces.append(
            RemoteTrace(
                name="receive_confirmed",
                extra_json=json.dumps({"num_subscribers": num_subscribers}),
                occurred_at_remote=time.time(),
            )
        )
        return self

    # HandleTrustedNotifyUnsubscribeImmediateDone
    def on_unsubscribe_immediate_success(
        self,
    ) -> HandledTrustedNotify[StatelessTracingNotifyOnSendingResponse]:
        self.traces.append(
            RemoteTrace(
                name="unsubscribe_immediate_success",
                extra_json="{}",
                occurred_at_remote=time.time(),
            )
        )
        return self

    def on_unsubscribe_immediate_not_found(
        self,
    ) -> HandledTrustedNotify[StatelessTracingNotifyOnSendingResponse]:
        self.traces.append(
            RemoteTrace(
                name="unsubscribe_immediate_not_found",
                extra_json="{}",
                occurred_at_remote=time.time(),
            )
        )
        return self

    def on_unsubscribe_immediate_unavailable(self) -> None:
        self.traces.append(
            RemoteTrace(
                name="unsubscribe_immediate_unavailable",
                extra_json="{}",
                occurred_at_remote=time.time(),
            )
        )

    # StatelessTracingNotifyOnSendingResponse
    def on_sending_response(self) -> bytes:
        assert self.request_received_at_nano is not None, "no request received"
        tracing = io.BytesIO()
        tracing.write(b"\0")  # truncated flag
        tracing.write(len(self.traces).to_bytes(2, "big"))

        for idx, trace in enumerate(self.traces):
            start_pos = tracing.tell()
            tracing.write(len(trace.name).to_bytes(2, "big"))
            tracing.write(trace.name.encode("utf-8"))
            tracing.write(len(trace.extra_json).to_bytes(2, "big"))
            tracing.write(trace.extra_json.encode("utf-8"))
            tracing.write(struct.pack(">d", trace.occurred_at_remote))
            if tracing.tell() >= (2**16) - 17:
                warnings.warn("tracing data too large, truncating")
                tracing.seek(0)
                tracing.write(b"\1")
                tracing.write(idx.to_bytes(2, "big"))
                tracing.seek(start_pos)
                break

        tracing.write(self.request_received_at_nano.to_bytes(8, "big"))
        response_sent_at_nano = time.time_ns()
        tracing.write(response_sent_at_nano.to_bytes(8, "big"))
        final_pos = tracing.tell()
        return tracing.getvalue()[:final_pos]


if TYPE_CHECKING:
    _: Type[StatelessTracingNotifyStart] = SimpleStatelessTracingNotifyStart



# ruff: noqa: T201
"""Tests that cancellation works as well as is possible and shows the
somewhat unintuitive but expected behavior, which also highlights why
wait_not_empty() combined with get_no_wait() is much easier to use 
compared to get() (and similarly, wait_not_full() with put_no_wait() is
easier to use compared to put()). Essentially, if cancellation goes
through depends on if there are event loop iterations after cancellation
"""

import asyncio

from lonelypsp.util.drainable_asyncio_queue import DrainableAsyncioQueue


async def _test() -> None:
    await _test_cancel_put_no_yield()
    print(".", end="", flush=True)
    await _test_cancel_put_with_yield()
    print(".", end="", flush=True)
    await _test_cancel_wait_not_full_no_yield()
    print(".", end="", flush=True)
    await _test_cancel_wait_not_full_with_yield()
    print(".", end="", flush=True)
    await _test_cancel_get_no_yield()
    print(".", end="", flush=True)
    await _test_cancel_get_with_yield()
    print(".", end="", flush=True)
    await _test_cancel_wait_not_empty_no_yield()
    print(".", end="", flush=True)
    await _test_cancel_wait_not_empty_with_yield()
    print(".", end="", flush=True)
    print("done")


async def _test_cancel_put_no_yield() -> None:
    q: DrainableAsyncioQueue[int] = DrainableAsyncioQueue(max_size=1)
    q.put_nowait(1)
    putter = asyncio.create_task(q.put(2))
    await asyncio.sleep(0)
    assert putter.cancel()  # does successfully schedule a CancelledError
    drained = q.drain()  # CancelledError hasn't been raised yet
    assert drained == [1, 2], f"{drained=}, expected [1, 2]"
    assert not putter.cancelled()  # will be able to see the cancellation hasn't run yet
    assert not putter.done()  # same as previous but for clarity
    await putter  # raises the CancelledError, detects finished normally (no error)


async def _test_cancel_put_with_yield() -> None:
    q: DrainableAsyncioQueue[int] = DrainableAsyncioQueue(max_size=1)
    q.put_nowait(1)
    putter = asyncio.create_task(q.put(2))
    await asyncio.sleep(0)
    assert putter.cancel()  # does successfully schedule a CancelledError
    await asyncio.sleep(0)  # raises the CancelledError
    drained = q.drain()
    assert drained == [1], f"{drained=}, expected [1]"
    assert putter.cancelled()  # will be able to see the cancellation has run
    assert putter.done()  # same as previous but for clarity


async def _test_cancel_wait_not_full_no_yield() -> None:
    q: DrainableAsyncioQueue[int] = DrainableAsyncioQueue(max_size=1)
    q.put_nowait(1)
    waiter = asyncio.create_task(q.wait_not_full())
    await asyncio.sleep(0)
    assert waiter.cancel()  # does successfully schedule a CancelledError
    assert q.drain() == [1]  # CancelledError hasn't been raised yet
    assert not waiter.cancelled()  # will be able to see the cancellation hasn't run yet
    assert not waiter.done()  # same as previous but for clarity
    await waiter  # raises the CancelledError, detects finished normally (no error)


async def _test_cancel_wait_not_full_with_yield() -> None:
    q: DrainableAsyncioQueue[int] = DrainableAsyncioQueue(max_size=1)
    q.put_nowait(1)
    waiter = asyncio.create_task(q.wait_not_full())
    await asyncio.sleep(0)
    assert waiter.cancel()  # does successfully schedule a CancelledError
    await asyncio.sleep(0)  # raises the CancelledError
    assert q.drain() == [1]
    assert waiter.cancelled()  # will be able to see the cancellation has run
    assert waiter.done()  # same as previous but for clarity


async def _test_cancel_get_no_yield() -> None:
    q: DrainableAsyncioQueue[int] = DrainableAsyncioQueue(max_size=1)
    getter = asyncio.create_task(q.get())
    await asyncio.sleep(0)
    assert getter.cancel()  # does successfully schedule a CancelledError
    q.put_nowait(1)  # CancelledError hasn't been raised yet
    assert not getter.cancelled()  # will be able to see the cancellation hasn't run yet
    assert not getter.done()  # same as previous but for clarity
    assert (
        await getter == 1
    )  # raises the CancelledError, detects finished normally (no error)
    drained = q.drain()
    assert drained == [], f"{drained=}, expected []"


async def _test_cancel_get_with_yield() -> None:
    q: DrainableAsyncioQueue[int] = DrainableAsyncioQueue(max_size=1)
    getter = asyncio.create_task(q.get())
    await asyncio.sleep(0)
    assert getter.cancel()  # does successfully schedule a CancelledError
    await asyncio.sleep(0)  # raises the CancelledError
    q.put_nowait(1)
    assert getter.cancelled()  # will be able to see the cancellation has run
    assert getter.done()  # same as previous but for clarity
    drained = q.drain()
    assert drained == [1], f"{drained=}, expected [1]"


async def _test_cancel_wait_not_empty_no_yield() -> None:
    q: DrainableAsyncioQueue[int] = DrainableAsyncioQueue(max_size=1)
    waiter = asyncio.create_task(q.wait_not_empty())
    await asyncio.sleep(0)
    assert waiter.cancel()  # does successfully schedule a CancelledError
    q.put_nowait(1)  # CancelledError hasn't been raised yet
    assert not waiter.cancelled()  # will be able to see the cancellation hasn't run yet
    assert not waiter.done()  # same as previous but for clarity
    await waiter  # raises the CancelledError, detects finished normally (no error)
    drained = q.drain()
    assert drained == [1], f"{drained=}, expected [1]"


async def _test_cancel_wait_not_empty_with_yield() -> None:
    q: DrainableAsyncioQueue[int] = DrainableAsyncioQueue(max_size=1)
    waiter = asyncio.create_task(q.wait_not_empty())
    await asyncio.sleep(0)
    assert waiter.cancel()  # does successfully schedule a CancelledError
    await asyncio.sleep(0)  # raises the CancelledError
    q.put_nowait(1)
    assert waiter.cancelled()  # will be able to see the cancellation has run
    assert waiter.done()  # same as previous but for clarity
    drained = q.drain()
    assert drained == [1], f"{drained=}, expected [1]"


if __name__ == "__main__":
    asyncio.run(_test())



from typing import TYPE_CHECKING, Collection, List, Literal, Optional, Type, Union

from lonelypsp.compat import fast_dataclass
from lonelypsp.stateful.constants import (
    PubSubStatefulMessageFlags,
    SubscriberToBroadcasterStatefulMessageType,
)
from lonelypsp.stateful.generic_parser import S2B_MessageParser
from lonelypsp.stateful.parser_helpers import parse_simple_headers
from lonelypsp.stateful.serializer_helpers import (
    MessageSerializer,
    serialize_simple_message,
)
from lonelypsp.sync_io import SyncReadableBytesIO


@fast_dataclass
class S2B_SubscribeExact:
    """
    S2B = Subscriber to Broadcaster
    See the type enum documentation for more information on the fields
    """

    type: Literal[SubscriberToBroadcasterStatefulMessageType.SUBSCRIBE_EXACT]
    """discriminator value"""

    authorization: Optional[str]
    """url: websocket:<nonce>:<ctr>
    
    an empty string is reinterpreted as None for consistency between
    minimal headers mode and expanded headers mode
    """

    tracing: bytes
    """the tracing data, which may be empty"""

    topic: bytes
    """the topic to subscribe to"""


_exact_headers: Collection[str] = ("authorization", "x-tracing", "x-topic")


class S2B_SubscribeExactParser:
    """Satisfies S2B_MessageParser[S2B_SubscribeExact]"""

    @classmethod
    def relevant_types(cls) -> List[SubscriberToBroadcasterStatefulMessageType]:
        return [SubscriberToBroadcasterStatefulMessageType.SUBSCRIBE_EXACT]

    @classmethod
    def parse(
        cls,
        flags: PubSubStatefulMessageFlags,
        type: SubscriberToBroadcasterStatefulMessageType,
        payload: SyncReadableBytesIO,
    ) -> S2B_SubscribeExact:
        assert type == SubscriberToBroadcasterStatefulMessageType.SUBSCRIBE_EXACT

        headers = parse_simple_headers(flags, payload, _exact_headers)

        authorization_bytes = headers["authorization"]
        authorization: Optional[str] = None
        if authorization_bytes != b"":
            try:
                authorization = authorization_bytes.decode("utf-8")
            except UnicodeDecodeError:
                raise ValueError("authorization must be a utf-8 string")

        tracing = headers["x-tracing"]

        topic = headers["x-topic"]
        return S2B_SubscribeExact(
            type=type,
            authorization=authorization,
            tracing=tracing,
            topic=topic,
        )


if TYPE_CHECKING:
    _: Type[S2B_MessageParser[S2B_SubscribeExact]] = S2B_SubscribeExactParser


def serialize_s2b_subscribe_exact(
    msg: S2B_SubscribeExact, /, *, minimal_headers: bool
) -> Union[bytes, bytearray]:
    """Satisfies MessageSerializer[S2B_SubscribeExact]"""
    return serialize_simple_message(
        type=msg.type,
        header_names=_exact_headers,
        header_values=(
            b"" if msg.authorization is None else msg.authorization.encode("utf-8"),
            msg.tracing,
            msg.topic,
        ),
        minimal_headers=minimal_headers,
        payload=b"",
    )


if TYPE_CHECKING:
    __: MessageSerializer[S2B_SubscribeExact] = serialize_s2b_subscribe_exact


@fast_dataclass
class S2B_SubscribeGlob:
    """
    S2B = Subscriber to Broadcaster
    See the type enum documentation for more information on the fields
    """

    type: Literal[SubscriberToBroadcasterStatefulMessageType.SUBSCRIBE_GLOB]
    """discriminator value"""

    authorization: Optional[str]
    """url: websocket:<nonce>:<ctr>
    
    an empty string is reinterpreted as None for consistency between
    minimal headers mode and expanded headers mode
    """

    tracing: bytes
    """the tracing data, which may be empty"""

    glob: str
    """the glob pattern to subscribe to"""


_glob_headers: Collection[str] = ("authorization", "x-tracing", "x-glob")


class S2B_SubscribeGlobParser:
    """Satisfies S2B_MessageParser[S2B_SubscribeGlob]"""

    @classmethod
    def relevant_types(cls) -> List[SubscriberToBroadcasterStatefulMessageType]:
        return [SubscriberToBroadcasterStatefulMessageType.SUBSCRIBE_GLOB]

    @classmethod
    def parse(
        cls,
        flags: PubSubStatefulMessageFlags,
        type: SubscriberToBroadcasterStatefulMessageType,
        payload: SyncReadableBytesIO,
    ) -> S2B_SubscribeGlob:
        assert type == SubscriberToBroadcasterStatefulMessageType.SUBSCRIBE_GLOB

        headers = parse_simple_headers(flags, payload, _glob_headers)

        authorization_bytes = headers["authorization"]
        authorization: Optional[str] = None
        if authorization_bytes != b"":
            try:
                authorization = authorization_bytes.decode("utf-8")
            except UnicodeDecodeError:
                raise ValueError("authorization must be a utf-8 string")

        tracing = headers["x-tracing"]

        glob = headers["x-glob"].decode("utf-8")
        return S2B_SubscribeGlob(
            type=type,
            authorization=authorization,
            tracing=tracing,
            glob=glob,
        )


if TYPE_CHECKING:
    ___: Type[S2B_MessageParser[S2B_SubscribeGlob]] = S2B_SubscribeGlobParser


def serialize_s2b_subscribe_glob(
    msg: S2B_SubscribeGlob, /, *, minimal_headers: bool
) -> Union[bytes, bytearray]:
    """Satisfies MessageSerializer[S2B_SubscribeGlob]"""
    return serialize_simple_message(
        type=msg.type,
        header_names=_glob_headers,
        header_values=(
            b"" if msg.authorization is None else msg.authorization.encode("utf-8"),
            msg.tracing,
            msg.glob.encode("utf-8"),
        ),
        minimal_headers=minimal_headers,
        payload=b"",
    )


if TYPE_CHECKING:
    ____: MessageSerializer[S2B_SubscribeGlob] = serialize_s2b_subscribe_glob

S2B_Subscribe = Union[S2B_SubscribeExact, S2B_SubscribeGlob]



from typing import Generic, List, Protocol, TypeVar

from lonelypsp.stateful.constants import (
    BroadcasterToSubscriberStatefulMessageType,
    PubSubStatefulMessageFlags,
    SubscriberToBroadcasterStatefulMessageType,
)
from lonelypsp.sync_io import SyncReadableBytesIO

T_co = TypeVar("T_co", covariant=True)


class S2B_MessageParser(Generic[T_co], Protocol):
    """Describes something that can parse a message from a subscriber to a broadcaster"""

    @classmethod
    def relevant_types(cls) -> List[SubscriberToBroadcasterStatefulMessageType]:
        """Returns the list of messages that this parser can parse"""
        ...

    @classmethod
    def parse(
        cls,
        flags: PubSubStatefulMessageFlags,
        type: SubscriberToBroadcasterStatefulMessageType,
        payload: SyncReadableBytesIO,
    ) -> T_co:
        """Parses a message from a subscriber to a broadcaster

        Args:
            flags (PubSubStatefulMessageFlags): the flags for the message
            type (SubscriberToBroadcasterStatefulMessageType): the type of the message
            payload (IO[bytes]): the payload of the message, seeked to after the
                type (i.e,. at the start of the payload)

        Returns:
            T: the parsed message

        Raises:
            AssertionError: if type is not in relevant_types
            ValueError: if the message is malformed
        """
        ...


class B2S_MessageParser(Generic[T_co], Protocol):
    """Describes something that can parse a message from a broadcaster to a subscriber"""

    @classmethod
    def relevant_types(cls) -> List[BroadcasterToSubscriberStatefulMessageType]:
        """Returns the list of messages that this parser can parse"""
        ...

    @classmethod
    def parse(
        cls,
        flags: PubSubStatefulMessageFlags,
        type: BroadcasterToSubscriberStatefulMessageType,
        payload: SyncReadableBytesIO,
    ) -> T_co:
        """Parses a message from a broadcaster to a subscriber

        Args:
            flags (PubSubStatefulMessageFlags): the flags for the message
            type (BroadcasterToSubscriberStatefulMessageType): the type of the message
            payload (readable file-like): the payload of the message, seeked to after the
                type (i.e,. at the start of the payload)

        Returns:
            T: the parsed message

        Raises:
            AssertionError: if type is not in relevant_types
            ValueError: if the message is malformed
        """
        ...



from typing import TYPE_CHECKING, Collection, List, Literal, Optional, Type, Union

from lonelypsp.compat import fast_dataclass
from lonelypsp.stateful.constants import (
    BroadcasterToSubscriberStatefulMessageType,
    PubSubStatefulMessageFlags,
)
from lonelypsp.stateful.generic_parser import B2S_MessageParser
from lonelypsp.stateful.parser_helpers import parse_simple_headers
from lonelypsp.stateful.serializer_helpers import (
    MessageSerializer,
    serialize_simple_message,
)
from lonelypsp.sync_io import SyncReadableBytesIO


@fast_dataclass
class B2S_ConfirmSubscribeExact:
    """
    B2S = Broadcaster to Subscriber
    See the type enum documentation for more information on the fields
    """

    type: Literal[BroadcasterToSubscriberStatefulMessageType.CONFIRM_SUBSCRIBE_EXACT]
    """discriminator value"""

    topic: bytes
    """the topic the subscriber is now subscribed to"""

    authorization: Optional[str]
    """the authorization header that shows the confirmation was sent by the broadcaster
    that was contacted
    
    empty strings are converted to None for consistency with http endpoints
    """

    tracing: bytes
    """the tracing data, which may be empty"""


@fast_dataclass
class B2S_ConfirmSubscribeGlob:
    """
    B2S = Broadcaster to Subscriber
    See the type enum documentation for more information on the fields
    """

    type: Literal[BroadcasterToSubscriberStatefulMessageType.CONFIRM_SUBSCRIBE_GLOB]
    """discriminator value"""

    glob: str
    """the glob pattern whose matching topics the subscriber is now subscribed to"""

    authorization: Optional[str]
    """the authorization header that shows the confirmation was sent by the broadcaster
    that was contacted
    
    empty strings are converted to None for consistency with http endpoints
    """

    tracing: bytes
    """the tracing data, which may be empty"""


B2S_ConfirmSubscribe = Union[B2S_ConfirmSubscribeExact, B2S_ConfirmSubscribeGlob]


_exact_headers: Collection[str] = ("x-topic", "authorization", "x-tracing")


class B2S_ConfirmSubscribeExactParser:
    """Satisfies B2S_MessageParser[B2S_ConfirmSubscribeExact]"""

    @classmethod
    def relevant_types(cls) -> List[BroadcasterToSubscriberStatefulMessageType]:
        return [BroadcasterToSubscriberStatefulMessageType.CONFIRM_SUBSCRIBE_EXACT]

    @classmethod
    def parse(
        cls,
        flags: PubSubStatefulMessageFlags,
        type: BroadcasterToSubscriberStatefulMessageType,
        payload: SyncReadableBytesIO,
    ) -> B2S_ConfirmSubscribeExact:
        assert (
            type == BroadcasterToSubscriberStatefulMessageType.CONFIRM_SUBSCRIBE_EXACT
        )

        headers = parse_simple_headers(flags, payload, _exact_headers)
        topic = headers["x-topic"]

        authorization_bytes = headers["authorization"]
        authorization: Optional[str] = None
        if authorization_bytes != b"":
            try:
                authorization = authorization_bytes.decode("utf-8")
            except UnicodeDecodeError:
                raise ValueError("authorization must be a utf-8 string")

        tracing = headers["x-tracing"]

        return B2S_ConfirmSubscribeExact(
            type=type,
            topic=topic,
            authorization=authorization,
            tracing=tracing,
        )


if TYPE_CHECKING:
    _: Type[B2S_MessageParser[B2S_ConfirmSubscribeExact]] = (
        B2S_ConfirmSubscribeExactParser
    )


def serialize_b2s_confirm_subscribe_exact(
    msg: B2S_ConfirmSubscribeExact, /, *, minimal_headers: bool
) -> Union[bytes, bytearray]:
    """Satisfies MessageSerializer[B2S_ConfirmSubscribeExact]"""
    return serialize_simple_message(
        type=msg.type,
        header_names=_exact_headers,
        header_values=(
            msg.topic,
            msg.authorization.encode("utf-8") if msg.authorization is not None else b"",
            msg.tracing,
        ),
        minimal_headers=minimal_headers,
        payload=b"",
    )


if TYPE_CHECKING:
    __: MessageSerializer[B2S_ConfirmSubscribeExact] = (
        serialize_b2s_confirm_subscribe_exact
    )


_glob_headers: Collection[str] = ("x-glob", "authorization", "x-tracing")


class B2S_ConfirmSubscribeGlobParser:
    """Satisfies B2S_MessageParser[B2S_ConfirmSubscribeGlob]"""

    @classmethod
    def relevant_types(cls) -> List[BroadcasterToSubscriberStatefulMessageType]:
        return [BroadcasterToSubscriberStatefulMessageType.CONFIRM_SUBSCRIBE_GLOB]

    @classmethod
    def parse(
        cls,
        flags: PubSubStatefulMessageFlags,
        type: BroadcasterToSubscriberStatefulMessageType,
        payload: SyncReadableBytesIO,
    ) -> B2S_ConfirmSubscribeGlob:
        assert type == BroadcasterToSubscriberStatefulMessageType.CONFIRM_SUBSCRIBE_GLOB

        headers = parse_simple_headers(flags, payload, _glob_headers)
        glob = headers["x-glob"].decode("utf-8")

        authorization_bytes = headers["authorization"]
        authorization: Optional[str] = None
        if authorization_bytes != b"":
            try:
                authorization = authorization_bytes.decode("utf-8")
            except UnicodeDecodeError:
                raise ValueError("authorization must be a utf-8 string")

        tracing = headers["x-tracing"]
        return B2S_ConfirmSubscribeGlob(
            type=type,
            glob=glob,
            authorization=authorization,
            tracing=tracing,
        )


if TYPE_CHECKING:
    ___: Type[B2S_MessageParser[B2S_ConfirmSubscribeGlob]] = (
        B2S_ConfirmSubscribeGlobParser
    )


def serialize_b2s_confirm_subscribe_glob(
    msg: B2S_ConfirmSubscribeGlob, /, *, minimal_headers: bool
) -> Union[bytes, bytearray]:
    """Satisfies MessageSerializer[B2S_ConfirmSubscribeGlob]"""
    return serialize_simple_message(
        type=msg.type,
        header_names=_glob_headers,
        header_values=(
            msg.glob.encode("utf-8"),
            msg.authorization.encode("utf-8") if msg.authorization is not None else b"",
            msg.tracing,
        ),
        minimal_headers=minimal_headers,
        payload=b"",
    )


if TYPE_CHECKING:
    ____: MessageSerializer[B2S_ConfirmSubscribeGlob] = (
        serialize_b2s_confirm_subscribe_glob
    )



from typing import TYPE_CHECKING, Collection, List, Literal, Optional, Type, Union

from lonelypsp.compat import fast_dataclass
from lonelypsp.stateful.constants import (
    BroadcasterToSubscriberStatefulMessageType,
    PubSubStatefulMessageFlags,
)
from lonelypsp.stateful.generic_parser import B2S_MessageParser
from lonelypsp.stateful.parser_helpers import parse_simple_headers
from lonelypsp.stateful.serializer_helpers import (
    MessageSerializer,
    int_to_minimal_unsigned,
    serialize_simple_message,
)
from lonelypsp.sync_io import SyncReadableBytesIO


@fast_dataclass
class B2S_ConfirmNotify:
    """
    B2S = Broadcaster to Subscriber
    See the type enum documentation for more information on the fields
    """

    type: Literal[BroadcasterToSubscriberStatefulMessageType.CONFIRM_NOTIFY]
    """discriminator value"""

    identifier: bytes
    """an arbitrary identifier for the notification assigned by the subscriber; max 64 bytes
    """

    subscribers: int
    """how many subscribers were successfully notified"""

    authorization: Optional[str]
    """the authorization header that shows the confirmation was sent by the broadcaster
    that was contacted
    
    empty strings are converted to None for consistency with http endpoints
    """

    tracing: bytes
    """the tracing data, which may be empty"""


_headers: Collection[str] = (
    "x-identifier",
    "x-subscribers",
    "authorization",
    "x-tracing",
)


class B2S_ConfirmNotifyParser:
    """Satisfies B2S_MessageParser[B2S_ConfirmNotify]"""

    @classmethod
    def relevant_types(cls) -> List[BroadcasterToSubscriberStatefulMessageType]:
        return [BroadcasterToSubscriberStatefulMessageType.CONFIRM_NOTIFY]

    @classmethod
    def parse(
        cls,
        flags: PubSubStatefulMessageFlags,
        type: BroadcasterToSubscriberStatefulMessageType,
        payload: SyncReadableBytesIO,
    ) -> B2S_ConfirmNotify:
        assert type == BroadcasterToSubscriberStatefulMessageType.CONFIRM_NOTIFY

        headers = parse_simple_headers(flags, payload, _headers)
        identifier = headers["x-identifier"]
        if len(identifier) > 64:
            raise ValueError("x-identifier must be at most 64 bytes")

        subscriber_bytes = headers["x-subscribers"]
        if len(subscriber_bytes) > 8:
            raise ValueError("x-subscribers must be at most 8 bytes")

        subscribers = int.from_bytes(subscriber_bytes, "big")

        authorization_bytes = headers["authorization"]
        authorization: Optional[str] = None
        if authorization_bytes != b"":
            try:
                authorization = authorization_bytes.decode("utf-8")
            except UnicodeDecodeError:
                raise ValueError("authorization must be a utf-8 string")

        tracing = headers["x-tracing"]

        return B2S_ConfirmNotify(
            type=type,
            identifier=identifier,
            subscribers=subscribers,
            authorization=authorization,
            tracing=tracing,
        )


if TYPE_CHECKING:
    _: Type[B2S_MessageParser[B2S_ConfirmNotify]] = B2S_ConfirmNotifyParser


def serialize_b2s_confirm_notify(
    msg: B2S_ConfirmNotify, /, *, minimal_headers: bool
) -> Union[bytes, bytearray]:
    """Satisfies MessageSerializer[B2S_ConfirmNotify]"""
    return serialize_simple_message(
        type=msg.type,
        header_names=_headers,
        header_values=(
            msg.identifier,
            int_to_minimal_unsigned(msg.subscribers),
            msg.authorization.encode("utf-8") if msg.authorization is not None else b"",
            msg.tracing,
        ),
        payload=b"",
        minimal_headers=minimal_headers,
    )


if TYPE_CHECKING:
    __: MessageSerializer[B2S_ConfirmNotify] = serialize_b2s_confirm_notify



import asyncio
import os
import queue
import sqlite3
import threading
import time
from enum import Enum, auto
from types import TracebackType
from typing import Any, List, Literal, Optional, Tuple, Type, Union, cast

from lonelypsp.compat import fast_dataclass
from lonelypsp.tracing.impl.simple.config import SimpleTracingConfig


class SimpleTracingDB:
    """Not process-safe or thread-safe, intended to be used in a dedicated thread"""

    def __init__(self, database: str) -> None:
        self.database: str = database
        """the path to where the database is stored, typically `traces.db`"""

        self.conn: Optional[sqlite3.Connection] = None
        """The connection to the database, if open"""

        self.cursor: Optional[sqlite3.Cursor] = None
        """the default cursor for the database, if open"""

    def __enter__(self) -> "SimpleTracingDB":
        conn = sqlite3.connect(self.database)
        try:
            self._setup(conn)

            cursor = conn.cursor()
            try:
                cursor.execute("PRAGMA foreign_keys = ON")
            except BaseException:
                cursor.close()
                raise
        except BaseException:
            conn.close()
            raise

        self.conn = conn
        self.cursor = cursor
        return self

    def __exit__(
        self,
        exc_type: Type[BaseException],
        exc_value: BaseException,
        traceback: TracebackType,
    ) -> None:
        try:
            if self.cursor is not None:
                self.cursor.close()
        finally:
            if self.conn is not None:
                self.conn.close()

    def _setup(self, conn: sqlite3.Connection) -> None:
        cursor = conn.cursor()
        try:
            cursor.execute("PRAGMA foreign_keys = ON")
            cursor.execute("BEGIN IMMEDIATE TRANSACTION")
            cursor.execute(
                "CREATE TABLE IF NOT EXISTS migrations (name TEXT PRIMARY KEY) WITHOUT ROWID"
            )
            cursor.execute("COMMIT TRANSACTION")

            current_file_path = __file__
            current_directory = os.path.dirname(current_file_path)
            migrations_directory = os.path.join(current_directory, "migrations")

            sorted_migrations = sorted(os.listdir(migrations_directory))

            for migration in sorted_migrations:
                if not migration.endswith(".sql"):
                    continue

                cursor.execute("BEGIN DEFERRED TRANSACTION")
                cursor.execute("SELECT 1 FROM migrations WHERE name = ?", (migration,))
                ran_migration = cursor.fetchone() is not None
                cursor.execute("COMMIT TRANSACTION")

                if ran_migration:
                    continue

                migration_path = os.path.join(migrations_directory, migration)
                with open(migration_path, "r") as migration_file:
                    migration_sql = migration_file.read()

                try:
                    cursor.executescript(migration_sql)
                except BaseException as e:
                    raise RuntimeError(f"Failed to run migration {migration}") from e

                cursor.execute("BEGIN IMMEDIATE TRANSACTION")
                cursor.execute("INSERT INTO migrations VALUES (?)", (migration,))
                cursor.execute("COMMIT TRANSACTION")
        finally:
            cursor.close()


class _ThreadQueueItemType(Enum):
    """The type of item in the queue"""

    SCRIPT = auto()
    """A script to execute"""

    STOP = auto()
    """A signal to stop the thread"""


@fast_dataclass
class _ThreadQueueItemScript:
    """A script to execute"""

    type: Literal[_ThreadQueueItemType.SCRIPT]
    """enum discriminator"""

    commands: List[Tuple[str, List[Any]]]
    """the commands to execute() in order"""


@fast_dataclass
class _ThreadQueueItemStop:
    """A signal to stop the thread"""

    type: Literal[_ThreadQueueItemType.STOP]
    """enum discriminator"""


_ThreadQueueItem = Union[
    _ThreadQueueItemScript,
    _ThreadQueueItemStop,
]


def _sweep_traces(
    config: SimpleTracingConfig, cursor: sqlite3.Cursor, now: float
) -> None:
    if config.trace_max_age_seconds is None:
        return

    cursor.execute("BEGIN IMMEDIATE TRANSACTION")
    cursor.execute(
        "DELETE FROM stateless_notifies WHERE created_at < ?",
        (now - config.trace_max_age_seconds,),
    )
    cursor.execute("COMMIT TRANSACTION")


def _thread_main(
    config: SimpleTracingConfig, database: str, q: queue.SimpleQueue
) -> None:
    with SimpleTracingDB(database) as db:
        assert db.cursor is not None

        next_sweep_at: Optional[float] = None
        if config.trace_max_age_seconds is not None:
            next_sweep_at = time.time() + config.trace_max_age_seconds

        while True:
            now = time.time()
            timeout: Optional[float] = None
            if (
                config.trace_max_age_seconds is not None
                and next_sweep_at is not None
                and now >= next_sweep_at
            ):
                _sweep_traces(config, db.cursor, now)
                next_sweep_at = now + config.trace_max_age_seconds

            if next_sweep_at is not None:
                timeout = next_sweep_at - now

            try:
                result = cast(_ThreadQueueItem, q.get(block=True, timeout=timeout))
            except queue.Empty:
                continue
            if result.type == _ThreadQueueItemType.STOP:
                break

            for command, args in result.commands:
                db.cursor.execute(command, args)


class SimpleTracingDBSidecar:
    """Starts a thread where the database is setup and allows deferring an entire sequence
    of operations to be run in the same thread

    Acts as an asynchronous context manager for starting/stopping the thread
    """

    def __init__(self, config: SimpleTracingConfig, database: str) -> None:
        self.config: SimpleTracingConfig = config
        """Various configuration options"""

        self.database: str = database
        """The path to the database or :memory: for an in-memory store"""

        self._lock: asyncio.Lock = asyncio.Lock()
        """Lock for interacting with _thread to ensure only one thread is started"""

        self._thread: Optional[threading.Thread] = None
        """The thread running the database"""

        self._queue: Optional[queue.SimpleQueue] = None
        """The queue to push tasks to the thread"""

    async def __aenter__(self) -> "SimpleTracingDBSidecar":
        async with self._lock:
            if self._thread is not None:
                raise RuntimeError("Thread already started")

            q: queue.SimpleQueue = queue.SimpleQueue()
            try:
                thread = threading.Thread(
                    target=_thread_main,
                    args=(self.database, q),
                    daemon=True,
                )

                await asyncio.get_running_loop().run_in_executor(None, thread.start)
            except BaseException:
                q.put(_ThreadQueueItemStop(type=_ThreadQueueItemType.STOP))
                raise

            self._thread = thread
            self._queue = q
            return self

    async def __aexit__(
        self,
        exc_type: Type[BaseException],
        exc_value: BaseException,
        traceback: TracebackType,
    ) -> None:
        async with self._lock:
            if self._thread is None:
                raise RuntimeError("Thread not started")
            if self._queue is None:
                raise RuntimeError("Queue not available")

            q = self._queue
            thread = self._thread
            self._queue = None
            self._thread = None

            q.put(_ThreadQueueItemStop(type=_ThreadQueueItemType.STOP))
            await asyncio.get_running_loop().run_in_executor(None, thread.join)

    def enqueue(self, commands: List[Tuple[str, List[Any]]]) -> None:
        assert self._queue is not None
        assert self._thread is not None
        assert self._thread.is_alive()
        self._queue.put(
            _ThreadQueueItemScript(type=_ThreadQueueItemType.SCRIPT, commands=commands)
        )



from types import TracebackType
from typing import Dict, Generator, Generic, Iterable, Iterator, Type, TypeVar

from lonelypsp.compat import fast_dataclass
from lonelypsp.stateful.constants import (
    BroadcasterToSubscriberStatefulMessageType,
    PubSubStatefulMessageFlags,
    SubscriberToBroadcasterStatefulMessageType,
)
from lonelypsp.sync_io import SyncReadableBytesIO


@fast_dataclass
class B2S_MessagePrefix:
    """
    B2S = Broadcaster to subscriber

    Describes the first 4 bytes of the message that are common to all B2S messages
    which are required to know how to parse the rest of the message
    """

    flags: PubSubStatefulMessageFlags
    """bit flags for parsing the message"""

    type: BroadcasterToSubscriberStatefulMessageType
    """enum value describing the contents of the message"""


def parse_b2s_message_prefix(body: SyncReadableBytesIO) -> B2S_MessagePrefix:
    """Consumes and interprets the first four bytes of a broadcaster to
    subscriber message. This is 2 bytes for flags and then 2 bytes for the
    message type discriminator.

    Raises ValueError if the message is too short to contain the prefix
    or is otherwise malformed

    Propagates errors from reading the body
    """
    flags_bytes = read_exact(body, 2)
    flags_int = int.from_bytes(flags_bytes, "big")
    flags = PubSubStatefulMessageFlags(flags_int)

    message_type_bytes = read_exact(body, 2)
    message_type_int = int.from_bytes(message_type_bytes, "big")
    message_type = BroadcasterToSubscriberStatefulMessageType(message_type_int)

    return B2S_MessagePrefix(flags, message_type)


@fast_dataclass
class S2B_MessagePrefix:
    """
    S2B = Subscriber to Broadcaster

    Describes the first 4 bytes of the message that are common to all S2B messages
    which are required to know how to parse the rest of the message
    """

    flags: PubSubStatefulMessageFlags
    """bit flags for parsing the message"""

    type: SubscriberToBroadcasterStatefulMessageType
    """enum value describing the contents of the message"""


def parse_s2b_message_prefix(body: SyncReadableBytesIO) -> S2B_MessagePrefix:
    """Consumes and interprets the first four bytes of a subscriber to
    broadcaster message. This is 2 bytes for flags and then 2 bytes for the
    message type discriminator.

    Raises ValueError if the message is too short to contain the prefix
    or is otherwise malformed

    Propagates errors from reading the body
    """
    flags_bytes = read_exact(body, 2)
    flags_int = int.from_bytes(flags_bytes, "big")
    flags = PubSubStatefulMessageFlags(flags_int)

    message_type_bytes = read_exact(body, 2)
    message_type_int = int.from_bytes(message_type_bytes, "big")
    message_type = SubscriberToBroadcasterStatefulMessageType(message_type_int)

    return S2B_MessagePrefix(flags, message_type)


T = TypeVar("T")


class GeneratorContextManager(Generic[T]):
    """Convenience class which makes working with generators easier by
    acting as a context manager that automatically closes the generator
    and provides the iterator
    """

    def __init__(self, gen: Generator[T, None, None]) -> None:
        self.gen = gen

    def __enter__(self) -> Iterator[T]:
        return self.gen.__iter__()

    def __exit__(
        self,
        exc_type: Type[BaseException],
        exc_value: BaseException,
        traceback: TracebackType,
    ) -> None:
        self.gen.close()


def read_exact(stream: SyncReadableBytesIO, n: int) -> bytes:
    """Reads exactly the indicated number of bytes, raising an exception if
    the stream ends before the bytes are read
    """
    data = stream.read(n)
    if len(data) != n:
        raise ValueError(f"underflow: expected {n} bytes, got {len(data)}")
    return data


def parse_minimal_message_headers(
    body: SyncReadableBytesIO,
) -> Generator[bytes, None, None]:
    """Consumes and parses minimal headers from the body, yielding
    after each one. Should be interrupted once all headers have been
    read. This is parsing the following format, where N is some value
    that the caller can determine

    - Repeat N:
        - 2 bytes (L): length of header value, big-endian encoded, unsigned
        - L bytes: header value

    Raises ValueError if the headers are malformed

    Propagates errors from reading the body
    """

    while True:
        header_length_bytes = read_exact(body, 2)
        header_length = int.from_bytes(header_length_bytes, "big")
        header_value = read_exact(body, header_length)
        yield header_value


def parse_expanded_headers(body: SyncReadableBytesIO) -> Dict[str, bytes]:
    """Parses the expanded headers at the given position in the body,
    returning a dictionary of header names to header values. The header
    names are ascii and lowercase, so they are unambiguous. When a header
    is repeated, the last value is used.

    - 2 bytes (N): number of headers, big-endian encoded, unsigned
    - REPEAT N:
        - 2 bytes (M): length of header name, big-endian encoded, unsigned
        - M bytes: header name, ascii-encoded
        - 2 bytes (L): length of header value, big-endian encoded, unsigned
        - L bytes: header value

    Raises ValueError if the headers are malformed

    Propagates errors from reading the body
    """
    header_count_bytes = read_exact(body, 2)
    header_count = int.from_bytes(header_count_bytes, "big")

    headers: Dict[str, bytes] = {}
    for _ in range(header_count):
        header_name_length_bytes = read_exact(body, 2)
        header_name_length = int.from_bytes(header_name_length_bytes, "big")
        header_name = read_exact(body, header_name_length).decode("ascii").lower()

        header_value_length_bytes = read_exact(body, 2)
        header_value_length = int.from_bytes(header_value_length_bytes, "big")
        header_value = read_exact(body, header_value_length)

        headers[header_name] = header_value

    return headers


def parse_simple_headers(
    flags: PubSubStatefulMessageFlags,
    body: SyncReadableBytesIO,
    minimal_headers: Iterable[str],
) -> Dict[str, bytes]:
    """Consumes and parses headers from the body, returning a dictionary
    of header names to header values. The header names are ascii and lowercase,
    so they are unambiguous. When a header is repeated, the last value is used.

    When minimal headers are provided, assumes they are exactly those specified in
    the list. If the headers are malformed, raises ValueError.

    Propagates errors from reading the body
    """
    if (flags & PubSubStatefulMessageFlags.MINIMAL_HEADERS) != 0:
        result: Dict[str, bytes] = {}
        with GeneratorContextManager(parse_minimal_message_headers(body)) as parser:
            for name in minimal_headers:
                result[name] = next(parser)
        return result

    return parse_expanded_headers(body)



from typing import TYPE_CHECKING, Collection, List, Literal, Optional, Type, Union

from lonelypsp.compat import fast_dataclass
from lonelypsp.stateful.constants import (
    BroadcasterToSubscriberStatefulMessageType,
    PubSubStatefulMessageFlags,
)
from lonelypsp.stateful.generic_parser import B2S_MessageParser
from lonelypsp.stateful.parser_helpers import parse_simple_headers
from lonelypsp.stateful.serializer_helpers import (
    MessageSerializer,
    serialize_simple_message,
)
from lonelypsp.sync_io import SyncReadableBytesIO


@fast_dataclass
class B2S_ConfirmConfigure:
    """
    B2S = Broadcaster to Subscriber
    See the type enum documentation for more information on the fields
    """

    type: Literal[BroadcasterToSubscriberStatefulMessageType.CONFIRM_CONFIGURE]
    """discriminator value"""

    broadcaster_nonce: bytes
    """32 random bytes representing the broadcasters contribution to the connection nonce"""

    authorization: Optional[str]
    """variable length proof that the broadcaster is allowed to serve the subscriber
    
    empty strings are converted to None for consistency with http endpoints
    """

    tracing: bytes
    """the tracing data, which may be empty"""


_headers: Collection[str] = ("x-broadcaster-nonce", "authorization", "x-tracing")


class B2S_ConfirmConfigureParser:
    """Satisfies B2S_MessageParser[B2S_ConfirmConfigure]"""

    @classmethod
    def relevant_types(cls) -> List[BroadcasterToSubscriberStatefulMessageType]:
        return [BroadcasterToSubscriberStatefulMessageType.CONFIRM_CONFIGURE]

    @classmethod
    def parse(
        cls,
        flags: PubSubStatefulMessageFlags,
        type: BroadcasterToSubscriberStatefulMessageType,
        payload: SyncReadableBytesIO,
    ) -> B2S_ConfirmConfigure:
        assert type == BroadcasterToSubscriberStatefulMessageType.CONFIRM_CONFIGURE

        headers = parse_simple_headers(flags, payload, _headers)
        broadcaster_nonce = headers["x-broadcaster-nonce"]
        if len(broadcaster_nonce) != 32:
            raise ValueError("x-broadcaster-nonce must be 32 bytes")

        authorization_bytes = headers["authorization"]
        authorization: Optional[str] = None
        if authorization_bytes != b"":
            try:
                authorization = authorization_bytes.decode("utf-8")
            except UnicodeDecodeError:
                raise ValueError("authorization must be a utf-8 string")

        tracing: bytes = headers["x-tracing"]

        return B2S_ConfirmConfigure(
            type=type,
            broadcaster_nonce=broadcaster_nonce,
            authorization=authorization,
            tracing=tracing,
        )


if TYPE_CHECKING:
    _: Type[B2S_MessageParser[B2S_ConfirmConfigure]] = B2S_ConfirmConfigureParser


def serialize_b2s_confirm_configure(
    msg: B2S_ConfirmConfigure, /, *, minimal_headers: bool
) -> Union[bytes, bytearray]:
    """Satisfies MessageSerializer[B2S_ConfirmConfigure]"""
    return serialize_simple_message(
        type=msg.type,
        header_names=_headers,
        header_values=(
            msg.broadcaster_nonce,
            msg.authorization.encode("utf-8") if msg.authorization is not None else b"",
            msg.tracing,
        ),
        payload=b"",
        minimal_headers=minimal_headers,
    )


if TYPE_CHECKING:
    __: MessageSerializer[B2S_ConfirmConfigure] = serialize_b2s_confirm_configure



from typing import Collection, Generic, Protocol, TypeVar, Union

from lonelypsp.stateful.constants import (
    BroadcasterToSubscriberStatefulMessageType,
    PubSubStatefulMessageFlags,
    SubscriberToBroadcasterStatefulMessageType,
)
from lonelypsp.sync_io import PreallocatedBytesIO, SyncWritableBytesIO


def serialize_prefix(
    out: SyncWritableBytesIO,
    type: Union[
        SubscriberToBroadcasterStatefulMessageType,
        BroadcasterToSubscriberStatefulMessageType,
    ],
    /,
    *,
    minimal_headers: bool,
) -> None:
    """Writes the message flags and the type of message to the output stream."""
    out.write(
        int.to_bytes(
            PubSubStatefulMessageFlags.MINIMAL_HEADERS if minimal_headers else 0,
            2,
            "big",
        )
    )
    out.write(int.to_bytes(type, 2, "big"))


def serialize_minimal_headers(
    out: SyncWritableBytesIO, values: Collection[bytes]
) -> None:
    """Writes the header values to the output stream, in minimal headers mode.

    Minimal headers does not get prefixed with the number of headers and the
    name of the headers is implicit from their order, thus this can be called
    multiple times to write multiple headers.
    """
    for value in values:
        out.write(int.to_bytes(len(value), 2, "big"))
        out.write(value)


def serialize_expanded_headers(
    out: SyncWritableBytesIO,
    header_names: Collection[str],
    header_values: Collection[bytes],
    /,
) -> None:
    """Writes the header values to the output stream, in expanded headers mode.

    Expanded headers get prefixed with the number of headers and the name of
    the headers, thus this can be called only once to write all headers.
    """
    assert len(header_names) == len(header_values)
    out.write(int.to_bytes(len(header_names), 2, "big"))
    for name, value in zip(header_names, header_values):
        out.write(int.to_bytes(len(name), 2, "big"))
        out.write(name.encode("ascii"))
        out.write(int.to_bytes(len(value), 2, "big"))
        out.write(value)


def serialize_simple_headers(
    out: SyncWritableBytesIO,
    header_names: Collection[str],
    header_values: Collection[bytes],
    /,
    *,
    minimal: bool,
) -> None:
    """Writes the given headers for the contents of a websocket message
    to the output stream in the appropriate mode.
    """

    if minimal:
        serialize_minimal_headers(out, header_values)
    else:
        serialize_expanded_headers(out, header_names, header_values)


def serialize_simple_message(
    *,
    type: Union[
        SubscriberToBroadcasterStatefulMessageType,
        BroadcasterToSubscriberStatefulMessageType,
    ],
    header_names: Collection[str],
    header_values: Collection[bytes],
    payload: bytes,
    minimal_headers: bool,
) -> Union[bytes, bytearray]:
    """Serializes the entire contents of a websocket message with the given
    type, headers, and payload, returning the bytes contents to send along
    the websocket
    """
    assert len(header_names) == len(header_values)
    total_size = (
        2  # flags
        + 2  # type
        + (  # headers
            2 * len(header_values) + sum(len(value) for value in header_values)
            if minimal_headers
            else 2
            + 4 * len(header_values)
            + sum(len(name.encode("ascii")) for name in header_names)
            + sum(len(value) for value in header_values)
        )
        + len(payload)
    )

    out = PreallocatedBytesIO(total_size)
    serialize_prefix(out, type, minimal_headers=minimal_headers)
    serialize_simple_headers(out, header_names, header_values, minimal=minimal_headers)
    out.write(payload)
    assert out.tell() == total_size
    return out.buffer


def int_to_minimal_unsigned(n: int) -> bytes:
    """Converts an integer to a minimal unsigned byte representation. In order
    to parse this the parser will need to know the length in bytes from the protocol,
    e.g., because it's a header value.

    Result is big-endian encoded
    """
    assert n >= 0
    bit_length = (n.bit_length() - 1) // 8 + 1
    return n.to_bytes(bit_length, "big")


T_contra = TypeVar("T_contra", contravariant=True)


class MessageSerializer(Generic[T_contra], Protocol):
    """The protocol the serializer functions should satisfy, to ensure we
    are being consistent
    """

    def __call__(
        self, msg: T_contra, /, *, minimal_headers: bool
    ) -> Union[bytes, bytearray]:
        """Serializes the given message into the bytes to send within a websocket
        bytes message.

        Args:
            msg: The message to serialize
            minimal_headers: True to use minimal headers, False to use expanded headers.
                Generally this is configured individually by broadcasters and subscribers,
                and only turned off as part of the process of updating the protocol before
                being turned back on when everything is on the same version.
        """



import hashlib
from typing import TYPE_CHECKING, Collection, List, Literal, Optional, Type, Union

from lonelypsp.compat import fast_dataclass
from lonelypsp.stateful.constants import (
    PubSubStatefulMessageFlags,
    SubscriberToBroadcasterStatefulMessageType,
)
from lonelypsp.stateful.generic_parser import S2B_MessageParser
from lonelypsp.stateful.parser_helpers import parse_simple_headers
from lonelypsp.stateful.serializer_helpers import (
    MessageSerializer,
    int_to_minimal_unsigned,
    serialize_simple_message,
)
from lonelypsp.sync_io import SyncReadableBytesIO


@fast_dataclass
class S2B_NotifyUncompressed:
    """
    S2B = Subscriber to Broadcaster
    See the type enum documentation for more information on the fields
    """

    type: Literal[SubscriberToBroadcasterStatefulMessageType.NOTIFY]
    """discriminator value"""

    authorization: Optional[str]
    """url: websocket:<nonce>:<ctr>
    
    an empty string is reinterpreted as None for consistency between
    minimal headers mode and expanded headers mode
    """

    tracing: bytes
    """the tracing data, which may be empty"""

    identifier: bytes
    """an arbitrary identifier for this notification assigned by the subscriber; max 64 bytes
    """

    compressor_id: Literal[None]
    """discriminator value. We reinterpret a compressor id of 0 to None as we cannot
    type "strictly positive integers" in python for the counterpart to `Literal[0]`
    """

    topic: bytes
    """the topic of the message"""

    verified_uncompressed_sha512: bytes
    """The sha512 hash of the uncompressed message, 64 bytes, verified"""

    uncompressed_message: bytes
    """The message in uncompressed form"""


@fast_dataclass
class S2B_NotifyCompressed:
    """
    S2B = Subscriber to Broadcaster
    See the type enum documentation for more information on the fields
    """

    type: Literal[SubscriberToBroadcasterStatefulMessageType.NOTIFY]
    """discriminator value"""

    authorization: Optional[str]
    """url: websocket:<nonce>:<ctr>
    
    an empty string is reinterpreted as None for consistency between
    minimal headers mode and expanded headers mode
    """

    tracing: bytes
    """the tracing data, which may be empty"""

    identifier: bytes
    """an arbitrary identifier for this notification assigned by the subscriber; max 64 bytes
    """

    compressor_id: int
    """the id of the compressor used to compress the message"""

    topic: bytes
    """the topic of the message"""

    verified_compressed_sha512: bytes
    """The sha512 hash of the compressed message, 64 bytes, verified"""

    compressed_message: bytes
    """The message in compressed form"""

    decompressed_length: int
    """The expected, but unverified, length of the message after decompression"""


S2B_Notify = Union[S2B_NotifyUncompressed, S2B_NotifyCompressed]


_headers: Collection[str] = (
    "authorization",
    "x-tracing",
    "x-identifier",
    "x-topic",
    "x-compressor",
    "x-compressed-length",
    "x-decompressed-length",
    "x-compressed-sha512",
)


class S2B_NotifyParser:
    """Satisfies S2B_MessageParser[S2B_Notify]"""

    @classmethod
    def relevant_types(cls) -> List[SubscriberToBroadcasterStatefulMessageType]:
        return [SubscriberToBroadcasterStatefulMessageType.NOTIFY]

    @classmethod
    def parse(
        cls,
        flags: PubSubStatefulMessageFlags,
        type: SubscriberToBroadcasterStatefulMessageType,
        payload: SyncReadableBytesIO,
    ) -> S2B_Notify:
        assert type == SubscriberToBroadcasterStatefulMessageType.NOTIFY

        headers = parse_simple_headers(flags, payload, _headers)

        authorization_bytes = headers["authorization"]
        authorization: Optional[str] = None
        if authorization_bytes != b"":
            try:
                authorization = authorization_bytes.decode("utf-8")
            except UnicodeDecodeError:
                raise ValueError("authorization must be a utf-8 string")

        tracing = headers["x-tracing"]

        identifier = headers["x-identifier"]
        if len(identifier) > 64:
            raise ValueError("x-identifier must be at most 64 bytes")

        topic = headers["x-topic"]
        compressor_id_bytes = headers["x-compressor"]
        if len(compressor_id_bytes) > 8:
            raise ValueError("x-compressor must be at most 8 bytes")

        compressor_id = int.from_bytes(compressor_id_bytes, "big")

        compressed_length_bytes = headers["x-compressed-length"]
        if len(compressed_length_bytes) > 8:
            raise ValueError("x-compressed-length must be at most 8 bytes")

        compressed_length = int.from_bytes(compressed_length_bytes, "big")

        decompressed_length_bytes = headers["x-decompressed-length"]
        if len(decompressed_length_bytes) > 8:
            raise ValueError("x-decompressed-length must be at most 8 bytes")

        decompressed_length = int.from_bytes(decompressed_length_bytes, "big")

        compressed_sha512 = headers["x-compressed-sha512"]
        if len(compressed_sha512) != 64:
            raise ValueError("x-compressed-sha512 must be 64 bytes")

        message = payload.read(-1)

        if len(message) != compressed_length:
            raise ValueError("x-compressed-length does not match the message length")

        message_digest = hashlib.sha512(message).digest()

        if message_digest != compressed_sha512:
            raise ValueError("x-compressed-sha512 does not match the message")

        if compressor_id == 0:
            if decompressed_length != compressed_length:
                raise ValueError(
                    "x-decompressed-length must equal x-compressed-length if x-compressor is 0"
                )

            return S2B_NotifyUncompressed(
                type=type,
                authorization=authorization,
                tracing=tracing,
                identifier=identifier,
                compressor_id=None,
                topic=topic,
                verified_uncompressed_sha512=compressed_sha512,
                uncompressed_message=message,
            )

        return S2B_NotifyCompressed(
            type=type,
            authorization=authorization,
            tracing=tracing,
            identifier=identifier,
            compressor_id=compressor_id,
            topic=topic,
            verified_compressed_sha512=compressed_sha512,
            compressed_message=message,
            decompressed_length=decompressed_length,
        )


if TYPE_CHECKING:
    _: Type[S2B_MessageParser[S2B_Notify]] = S2B_NotifyParser


def serialize_s2b_notify(
    msg: S2B_Notify, /, *, minimal_headers: bool
) -> Union[bytes, bytearray]:
    """Satisfies MessageSerializer[S2B_Notify]"""
    authorization_bytes = (
        b"" if msg.authorization is None else msg.authorization.encode("utf-8")
    )

    if msg.compressor_id is None:
        return serialize_simple_message(
            type=msg.type,
            header_names=_headers,
            header_values=(
                authorization_bytes,
                msg.identifier,
                msg.topic,
                b"\x00",
                int_to_minimal_unsigned(len(msg.uncompressed_message)),
                int_to_minimal_unsigned(len(msg.uncompressed_message)),
                msg.verified_uncompressed_sha512,
            ),
            payload=msg.uncompressed_message,
            minimal_headers=minimal_headers,
        )

    return serialize_simple_message(
        type=msg.type,
        header_names=_headers,
        header_values=(
            authorization_bytes,
            msg.identifier,
            msg.topic,
            int_to_minimal_unsigned(msg.compressor_id),
            int_to_minimal_unsigned(len(msg.compressed_message)),
            int_to_minimal_unsigned(msg.decompressed_length),
            msg.verified_compressed_sha512,
        ),
        payload=msg.compressed_message,
        minimal_headers=minimal_headers,
    )


if TYPE_CHECKING:
    __: MessageSerializer[S2B_Notify] = serialize_s2b_notify



import asyncio
from enum import Enum, auto
from types import TracebackType
from typing import (
    TYPE_CHECKING,
    Generic,
    Iterable,
    List,
    Literal,
    Optional,
    Type,
    TypeVar,
    Union,
)

from lonelypsp.compat import fast_dataclass
from lonelypsp.util.async_queue_like import AsyncQueueLike
from lonelypsp.util.bounded_deque import BoundedDeque

T = TypeVar("T")


class QueueDrained(Exception):
    """The exception raised when a drained queue is attempted to be accessed"""


class _GetType(Enum):
    GET = auto()
    WAIT = auto()


@fast_dataclass
class _GetterGet(Generic[T]):
    type: Literal[_GetType.GET]
    future: asyncio.Future[T]


@fast_dataclass
class _GetterWait:
    type: Literal[_GetType.WAIT]
    future: asyncio.Future[None]


class _PutType(Enum):
    PUT = auto()
    WAIT = auto()


@fast_dataclass
class _PutterPut(Generic[T]):
    type: Literal[_PutType.PUT]
    item: T
    future: asyncio.Future[None]


@fast_dataclass
class _PutterWait:
    type: Literal[_PutType.WAIT]
    future: asyncio.Future[None]


class DrainableAsyncioQueue(Generic[T]):
    """Satisfies AsyncioQueueLike[T] but adds the following functionality:

    - max size of 0 means that the queue is always full instead of an unbounded
      queue. a value of None is now also allowed and means an unbounded queue

    - adds optional maximum number of pending get() tasks when the queue is empty

    - adds optional maximum number of pending put() tasks when the queue is full

    - drain() method which empties the queue and raises a QueueDrained exception
      for any future get() or put() calls

    - wait_not_empty() and wait_not_full() methods which are non-consuming version of get/put
      which are safely cancelable at the cost of being inefficient if there are multiple
      consumers/producers respectively.

    - can be used as an asynchronous context manager, which calls drain() when exiting
    """

    def __init__(
        self,
        items: Optional[Iterable[T]] = None,
        /,
        *,
        max_size: Optional[int] = None,
        max_getters: Optional[int] = None,
        max_putters: Optional[int] = None,
    ) -> None:
        self._items: BoundedDeque[T] = BoundedDeque(maxlen=max_size)
        self._getters: BoundedDeque[Union[_GetterGet[T], _GetterWait]] = BoundedDeque(
            maxlen=max_getters
        )
        self._putters: BoundedDeque[Union[_PutterPut[T], _PutterWait]] = BoundedDeque(
            maxlen=max_putters
        )
        # after draining _items will have maxlen 0, but we raise a different
        # error
        self._drained = False

        if items is not None:
            for item in items:
                self._items.append(item)

    async def __aenter__(self) -> "DrainableAsyncioQueue[T]":
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self.drain()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({list(self._items)!r})"

    def __str__(self) -> str:
        return f"{{{self.__class__.__name__}; items=({self._items}) drained={self._drained} getters={len(self._getters)} putters={len(self._putters)}}}"

    def _on_put_one(self) -> bool:
        """Internal function to implement _on_put and _on_get; alerts the next
        getter if there is one and returns True, otherwise returns False

        Note that if there is a getter then we may now have space for another
        putter
        """
        while True:
            if not self._getters:
                return False

            getter = self._getters.popleft()
            if getter.type == _GetType.WAIT:
                getter.future.set_result(None)
                continue

            item = self._items.popleft()
            getter.future.set_result(item)
            return True

    def _on_get_one(self) -> bool:
        """Internal function to implement _on_put and _on_get; alerts the next
        putter if there is one and returns True, otherwise returns False

        Note that if there is a putter then we may now have space for another
        getter
        """
        while True:
            if not self._putters:
                return False

            putter = self._putters.popleft()
            if putter.type == _PutType.WAIT:
                putter.future.set_result(None)
                continue

            putter.future.set_result(None)
            self._items.append(putter.item)
            return True

    def _on_put(self) -> None:
        """Internal function called when a new item was put into the queue"""

        # utilizing short circuiting
        while self._on_put_one() and self._on_get_one():
            pass

    def _on_get(self) -> None:
        """Internal function called when an item was gotten from the queue"""

        # utilizing short circuiting
        while self._on_get_one() and self._on_put_one():
            pass

    def qsize(self) -> int:
        return len(self._items)

    def empty(self) -> bool:
        return not self._items

    def full(self) -> bool:
        return len(self._items) == self._items.max_size

    async def put(self, item: T) -> None:
        if not self.full():
            self._items.append(item)
            self._on_put()
            return

        if self._drained:
            raise QueueDrained

        putter: _PutterPut[T] = _PutterPut(
            type=_PutType.PUT, item=item, future=asyncio.Future()
        )
        self._putters.append(putter)
        try:
            await asyncio.wait([putter.future])
        except asyncio.CancelledError:
            if putter.future.cancel():
                self._putters.remove(putter)
                raise
            putter.future.result()

    def put_nowait(self, item: T) -> None:
        if not self.full():
            self._items.append(item)
            self._on_put()
            return

        if self._drained:
            raise QueueDrained

        raise asyncio.QueueFull

    async def get(self) -> T:
        if not self.empty():
            result = self._items.popleft()
            self._on_get()
            return result

        if self._drained:
            raise QueueDrained

        getter: _GetterGet[T] = _GetterGet(type=_GetType.GET, future=asyncio.Future())
        self._getters.append(getter)
        try:
            await asyncio.wait([getter.future])
        except asyncio.CancelledError:
            if getter.future.cancel():
                self._getters.remove(getter)
                raise
            return await getter.future

        return getter.future.result()

    def get_nowait(self) -> T:
        if not self.empty():
            result = self._items.popleft()
            self._on_get()
            return result

        if self._drained:
            raise QueueDrained

        raise asyncio.QueueEmpty

    async def wait_not_empty(self) -> None:
        """Between this entering and the returned coroutine returning normally, at some
        point the queue will have had an item available for get() call

        This is generally helpful iff there is a single consumer; if there are multiple
        consumers, between the caller being scheduled and being run the item may have already
        been consumed, leading to wasted cycles. However, a big advantage of this is it can
        be freely canceled without having to worry about if it completed or not

        Raises QueueDrained if the queue has been drained before this call or before an item
        is added
        """
        if not self.empty():
            return

        if self._drained:
            raise QueueDrained

        getter = _GetterWait(type=_GetType.WAIT, future=asyncio.Future())
        self._getters.append(getter)
        try:
            await asyncio.wait([getter.future])
        except asyncio.CancelledError:
            if getter.future.cancel():
                self._getters.remove(getter)
                raise
            await getter.future

    async def wait_not_full(self) -> None:
        """Between this entering and the returned coroutine returning normally, at some
        point the queue will have had space available for a put() call

        This is generally helpful iff there is a single producer; if there are multiple
        producers, between the caller being scheduled and being run the queue may have already
        been filled, leading to wasted cycles. However, a big advantage of this is it can
        be freely canceled without having to worry about if it completed or not

        Raises QueueDrained if the queue has been drained before this call, will complete
        normally if drained after being put into the queue
        """
        if not self.full():
            return

        if self._drained:
            raise QueueDrained

        putter = _PutterWait(type=_PutType.WAIT, future=asyncio.Future())
        self._putters.append(putter)
        try:
            await asyncio.wait([putter.future])
        except asyncio.CancelledError:
            if putter.future.cancel():
                self._putters.remove(putter)
                raise
            await putter.future

    def drain(self) -> List[T]:
        """Drains out the queue.

        If there are no items in the queue, causes any pending get calls
        to raise QueueDrained, then causes all future get or put calls to
        raise QueueDrained and returns an empty list.

        If there are items in the queue, moves those items into a list, allowing
        all pending put calls to be resolved (with their items included in the
        result), then causes all future get or put calls to raise QueueDrained
        and returns the list of items.
        """
        if self._drained:
            return []

        result = []
        while True:
            try:
                result.append(self.get_nowait())
            except asyncio.QueueEmpty:
                break

        self._drained = True
        for getter in self._getters:
            getter.future.set_exception(QueueDrained)
        assert not self._putters

        self._items = BoundedDeque(maxlen=0)
        self._getters = BoundedDeque(maxlen=0)
        self._putters = BoundedDeque(maxlen=0)
        return result


if TYPE_CHECKING:
    _: Type[AsyncQueueLike] = DrainableAsyncioQueue
